# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core import (
    SortComponents,
    Constraint,
    Objective,
    LogicalConstraint,
    Expression,
)
from pyomo.core.base import TransformationFactory, Suffix, ConstraintList, Integers
from pyomo.core.base.block import Block, TraversalStrategy
from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.common.modeling import unique_component_name
from pyomo.contrib.gdpopt.discrete_problem_initialize import valid_init_strategies
from pyomo.contrib.gdpopt.util import (
    get_main_elapsed_time,
    move_nonlinear_objective_to_constraints,
)
from pyomo.gdp.disjunct import Disjunct, Disjunction
from pyomo.util.vars_from_expressions import get_vars_from_components


def _get_discrete_problem_and_subproblem(solver, config):
    util_block = solver.original_util_block
    original_model = util_block.parent_block()
    if config.force_subproblem_nlp:
        # We'll need to fix these too
        add_discrete_variable_list(util_block)
    original_obj = move_nonlinear_objective_to_constraints(util_block, config.logger)
    solver.original_obj = original_obj

    # create model to hold the subproblems: We create this first because
    # certain initialization strategies for the discrete problem need it.
    subproblem, subproblem_util_block = get_subproblem(original_model, util_block)

    # create discrete problem--the MILP relaxation
    start = get_main_elapsed_time(solver.timing)
    discrete_problem_util_block = initialize_discrete_problem(
        util_block, subproblem_util_block, config, solver
    )

    config.logger.info(
        'Finished discrete problem initialization in {:.2f}s '
        'and {} iterations \n'.format(
            get_main_elapsed_time(solver.timing) - start,
            solver.initialization_iteration,
        )
    )

    return (discrete_problem_util_block, subproblem_util_block)


def initialize_discrete_problem(util_block, subprob_util_block, config, solver):
    """
    Calls the specified transformation (by default bigm) on the original
    model and removes nonlinear constraints to create a MILP discrete problem.
    """
    config.logger.info("---Starting GDPopt initialization---")
    # clone the original model
    discrete = util_block.parent_block().clone()
    discrete.name = discrete.name + ": discrete problem"

    discrete_problem_util_block = discrete.component(util_block.local_name)
    discrete_problem_util_block.no_good_cuts = ConstraintList()
    discrete_problem_util_block.no_good_disjunctions = Disjunction(Integers)

    # deactivate nonlinear constraints
    for c in discrete.component_data_objects(
        Constraint, active=True, descend_into=(Block, Disjunct)
    ):
        if c.body.polynomial_degree() not in (1, 0):
            c.deactivate()

    # Transform to a MILP
    TransformationFactory(config.discrete_problem_transformation).apply_to(discrete)
    add_transformed_boolean_variable_list(discrete_problem_util_block)
    add_algebraic_variable_list(discrete_problem_util_block, name='all_mip_variables')

    # Call the specified initialization strategy. (We've already validated the
    # input in the config logic, so we know this is okay.)
    init_algorithm = valid_init_strategies.get(config.init_algorithm)
    init_algorithm(
        util_block, discrete_problem_util_block, subprob_util_block, config, solver
    )

    return discrete_problem_util_block


def add_util_block(discrete):
    # create a block to store the cuts
    name = unique_component_name(discrete, '_gdpopt_cuts')
    block = Block()
    discrete.add_component(name, block)

    return block


def add_disjunct_list(util_block):
    model = util_block.parent_block()
    util_block.disjunct_list = list(
        model.component_data_objects(
            ctype=Disjunct,
            active=True,
            descend_into=(Block, Disjunct),
            sort=SortComponents.deterministic,
        )
    )


def add_disjunction_list(util_block):
    model = util_block.parent_block()
    util_block.disjunction_list = list(
        model.component_data_objects(
            ctype=Disjunction,
            active=True,
            descend_into=(Block, Disjunct),
            sort=SortComponents.deterministic,
        )
    )


def add_constraint_list(util_block):
    model = util_block.parent_block()
    util_block.constraint_list = list(
        model.component_data_objects(
            ctype=Constraint,
            active=True,
            descend_into=(Block, Disjunct),
            sort=SortComponents.deterministic,
        )
    )


def add_global_constraint_list(util_block):
    model = util_block.parent_block()
    util_block.global_constraint_list = list(
        model.component_data_objects(
            ctype=Constraint,
            active=True,
            descend_into=Block,
            sort=SortComponents.deterministic,
        )
    )


def add_constraints_by_disjunct(util_block):
    constraints_by_disjunct = util_block.constraints_by_disjunct = {}
    for disj in util_block.disjunct_list:
        cons_list = constraints_by_disjunct[disj] = []
        for cons in disj.component_data_objects(
            Constraint,
            active=True,
            descend_into=Block,
            sort=SortComponents.deterministic,
        ):
            cons_list.append(cons)


def add_algebraic_variable_list(util_block, name=None):
    """
    This collects variables from active Constraints and Objectives. It descends
    into Disjuncts, but does not collect any indicator variables that do not
    appear in algebraic constraints pre-transformation.
    """
    model = util_block.parent_block()
    if name is None:
        name = "algebraic_variable_list"
    setattr(
        util_block,
        name,
        list(
            get_vars_from_components(
                model,
                ctype=(Constraint, Objective),
                descend_into=(Block, Disjunct),
                active=True,
                sort=SortComponents.deterministic,
            )
        ),
    )


def add_discrete_variable_list(util_block):
    lst = util_block.discrete_variable_list = []
    for v in util_block.algebraic_variable_list:
        if v.is_integer():
            lst.append(v)


# Must be collected after list of Disjuncts
def add_boolean_variable_lists(util_block):
    util_block.boolean_variable_list = []
    util_block.non_indicator_boolean_variable_list = []
    for disjunct in util_block.disjunct_list:
        util_block.boolean_variable_list.append(disjunct.indicator_var)
    ind_var_set = ComponentSet(util_block.boolean_variable_list)
    # This will not necessarily include the indicator_vars if it is called
    # before the GDP is transformed to a MIP.
    for v in get_vars_from_components(
        util_block.parent_block(),
        ctype=LogicalConstraint,
        descend_into=(Block, Disjunct),
        active=True,
        sort=SortComponents.deterministic,
    ):
        if v not in ind_var_set:
            util_block.boolean_variable_list.append(v)
            util_block.non_indicator_boolean_variable_list.append(v)


# For the discrete problem, we want the corresponding binaries for all of the
# BooleanVars. This must be called after logical_to_disjunctive has been called.
def add_transformed_boolean_variable_list(util_block):
    util_block.transformed_boolean_variable_list = [
        v.get_associated_binary() for v in util_block.boolean_variable_list
    ]


def get_subproblem(original_model, util_block):
    """Clone the original, and reclassify all the Disjuncts to Blocks.
    We'll also call logical_to_disjunctive and bigm the disjunctive parts in
    case any of the indicator_vars are used in logical constraints and to make
    sure that the rest of the model is algebraic (assuming it was a proper
    GDP to begin with).
    """
    subproblem = original_model.clone()
    subproblem.name = subproblem.name + ": subproblem"

    # Set up dual value reporting
    if not hasattr(subproblem, 'dual'):
        subproblem.dual = Suffix(direction=Suffix.IMPORT)
    elif not isinstance(subproblem.dual, Suffix):
        raise ValueError(
            "The model contains a component called 'dual' that "
            "is not a Suffix. It is of type %s. Please rename "
            "this component, as GDPopt needs dual information to "
            "create cuts." % type(subproblem.dual)
        )
    subproblem.dual.activate()

    # reclassify all the Disjuncts as Blocks and deactivate the Disjunctions. We
    # don't need to add the xor constraints because we're not going to pass
    # infeasible integer solutions to this model.
    for disjunction in subproblem.component_data_objects(
        Disjunction,
        descend_into=(Block, Disjunct),
        descent_order=TraversalStrategy.PostfixDFS,
    ):
        for disjunct in disjunction.disjuncts:
            if disjunct.indicator_var.fixed:
                if not disjunct.indicator_var.value:
                    disjunct.deactivate()
            disjunct.parent_block().reclassify_component_type(disjunct, Block)

        disjunction.deactivate()

    TransformationFactory('contrib.logical_to_disjunctive').apply_to(subproblem)
    # transform any of the Disjuncts we created above with bigm.
    TransformationFactory('gdp.bigm').apply_to(subproblem)

    subproblem_util_block = subproblem.component(util_block.local_name)
    save_initial_values(subproblem_util_block)
    add_transformed_boolean_variable_list(subproblem_util_block)
    subproblem_obj = next(
        subproblem.component_data_objects(Objective, active=True, descend_into=True)
    )
    subproblem_util_block.obj = Expression(expr=subproblem_obj.expr)

    return subproblem, subproblem_util_block


def save_initial_values(subproblem_util_block):
    initial_values = subproblem_util_block.initial_var_values = ComponentMap()
    for v in subproblem_util_block.algebraic_variable_list:
        initial_values[v] = v.value



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.autoslots import AutoSlots


# Note: in an ideal world, PyomoObject would use the AutoSlots
# metaclass.  However, declaring a custom (non-type) metaclass has
# measurable performance implications.  It is faster to just look for
# the __auto_slots__ attribute and generate it if it is not present than
# to slow down the entire class hierarchy by declaring a metaclass.
class PyomoObject(AutoSlots.Mixin):
    __slots__ = ()

    def is_component_type(self):
        """Return True if this class is a Pyomo component"""
        return False

    def is_numeric_type(self):
        """Return True if this class is a Pyomo numeric object"""
        return False

    def is_parameter_type(self):
        """Return False unless this class is a parameter object"""
        return False

    def is_variable_type(self):
        """Return False unless this class is a variable object"""
        return False

    def is_expression_type(self, expression_system=None):
        """Return True if this numeric value is an expression"""
        return False

    def is_named_expression_type(self):
        """Return True if this numeric value is a named expression"""
        return False

    def is_logical_type(self):
        """Return True if this class is a Pyomo Boolean object.

        Boolean objects include constants, variables, or logical expressions.
        """
        return False

    def is_reference(self):
        """Return True if this object is a reference."""
        return False



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# ScenariosCreator.py - Class to create and deliver scenarios using parmest
# DLW March 2020

import pyomo.environ as pyo

import logging

logger = logging.getLogger(__name__)


class ScenarioSet:
    """
    Class to hold scenario sets

    Args:
    name (str): name of the set (might be "")

    """

    def __init__(self, name):
        # Note: If there was a use-case, the list could be a dataframe.
        self._scens = list()  # use a df instead?
        self.name = name  #  might be ""

    def _firstscen(self):
        # Return the first scenario for testing and to get Theta names.
        assert len(self._scens) > 0
        return self._scens[0]

    def ScensIterator(self):
        """Usage: for scenario in ScensIterator()"""
        return iter(self._scens)

    def ScenarioNumber(self, scennum):
        """Returns the scenario with the given, zero-based number"""
        return self._scens[scennum]

    def addone(self, scen):
        """Add a scenario to the set

        Args:
            scen (ParmestScen): the scenario to add
        """
        assert isinstance(self._scens, list)
        self._scens.append(scen)

    def append_bootstrap(self, bootstrap_theta):
        """Append a bootstrap theta df to the scenario set; equally likely

        Args:
            bootstrap_theta (dataframe): created by the bootstrap
        Note: this can be cleaned up a lot with the list becomes a df,
              which is why I put it in the ScenarioSet class.
        """
        assert len(bootstrap_theta) > 0
        prob = 1.0 / len(bootstrap_theta)

        # dict of ThetaVal dicts
        dfdict = bootstrap_theta.to_dict(orient='index')

        for index, ThetaVals in dfdict.items():
            name = "Bootstrap" + str(index)
            self.addone(ParmestScen(name, ThetaVals, prob))

    def write_csv(self, filename):
        """write a csv file with the scenarios in the set

        Args:
            filename (str): full path and full name of file
        """
        if len(self._scens) == 0:
            print("Empty scenario set, not writing file={}".format(filename))
            return
        with open(filename, "w") as f:
            f.write("Name,Probability")
            for n in self._firstscen().ThetaVals.keys():
                f.write(",{}".format(n))
            f.write('\n')
            for s in self.ScensIterator():
                f.write("{},{}".format(s.name, s.probability))
                for v in s.ThetaVals.values():
                    f.write(",{}".format(v))
                f.write('\n')


class ParmestScen:
    """A little container for scenarios; the Args are the attributes.

    Args:
        name (str): name for reporting; might be ""
        ThetaVals (dict): ThetaVals[name]=val
        probability (float): probability of occurrence "near" these ThetaVals
    """

    def __init__(self, name, ThetaVals, probability):
        self.name = name
        assert isinstance(ThetaVals, dict)
        self.ThetaVals = ThetaVals
        self.probability = probability


############################################################


class ScenarioCreator:
    """Create scenarios from parmest.

    Args:
        pest (Estimator): the parmest object
        solvername (str): name of the solver (e.g. "ipopt")

    """

    def __init__(self, pest, solvername):

        self.pest = pest
        self.solvername = solvername

    def ScenariosFromExperiments(self, addtoSet):
        """Creates new self.Scenarios list using the experiments only.

        Args:
            addtoSet (ScenarioSet): the scenarios will be added to this set
        Returns:
            a ScenarioSet
        """

        assert isinstance(addtoSet, ScenarioSet)

        if self.pest.pest_deprecated is not None:
            scenario_numbers = list(range(len(self.pest.pest_deprecated.callback_data)))
        else:
            scenario_numbers = list(range(len(self.pest.exp_list)))

        prob = 1.0 / len(scenario_numbers)
        for exp_num in scenario_numbers:
            ##print("Experiment number=", exp_num)
            if self.pest.pest_deprecated is not None:
                model = self.pest.pest_deprecated._instance_creation_callback(
                    exp_num, self.pest.pest_deprecated.callback_data
                )
            else:
                model = self.pest._instance_creation_callback(exp_num)
            opt = pyo.SolverFactory(self.solvername)
            results = opt.solve(model)  # solves and updates model
            ## pyo.check_termination_optimal(results)
            if self.pest.pest_deprecated is not None:
                ThetaVals = {
                    theta: pyo.value(model.find_component(theta))
                    for theta in self.pest.pest_deprecated.theta_names
                }
            else:
                ThetaVals = {
                    k.name: pyo.value(k) for k in model.unknown_parameters.keys()
                }
            addtoSet.addone(ParmestScen("ExpScen" + str(exp_num), ThetaVals, prob))

    def ScenariosFromBootstrap(self, addtoSet, numtomake, seed=None):
        """Creates new self.Scenarios list using the experiments only.

        Args:
            addtoSet (ScenarioSet): the scenarios will be added to this set
            numtomake (int) : number of scenarios to create
        """

        assert isinstance(addtoSet, ScenarioSet)

        if self.pest.pest_deprecated is not None:
            bootstrap_thetas = self.pest.pest_deprecated.theta_est_bootstrap(
                numtomake, seed=seed
            )
        else:
            bootstrap_thetas = self.pest.theta_est_bootstrap(numtomake, seed=seed)
        addtoSet.append_bootstrap(bootstrap_thetas)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# the purpose of this file is to collect all utility methods that compute
# attributes of blocks, based on their contents.

from pyomo.common import deprecated
from pyomo.core.base import Var


@deprecated("This function has been moved to `pyomo.util.blockutil`", version='5.6.9')
def has_discrete_variables(block):
    from pyomo.util.blockutil import has_discrete_variables

    return has_discrete_variables(block)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import collections.abc

from pyomo.core.kernel.homogeneous_container import IHomogeneousContainer

logger = logging.getLogger('pyomo.core')


class DictContainer(IHomogeneousContainer, collections.abc.MutableMapping):
    """
    A partial implementation of the IHomogeneousContainer
    interface that provides dict-like storage functionality.

    Complete implementations need to set the _ctype property
    at the class level and initialize the remaining
    ICategorizedObject attributes during object creation. If
    using __slots__, a slot named "_data" must be included.

    Note that this implementation allows nested storage of
    other ICategorizedObjectContainer implementations that
    are defined with the same ctype.
    """

    __slots__ = ()
    _child_storage_delimiter_string = ""
    _child_storage_entry_string = "[%s]"

    def __init__(self, *args, **kwds):
        self._data = dict()
        if len(args) > 0:
            if len(args) > 1:
                raise TypeError(
                    "%s expected at most 1 arguments, "
                    "got %s" % (self.__class__.__name__, len(args))
                )
            self.update(args[0])
        if len(kwds):
            self.update(**kwds)

    def _fast_insert(self, key, item):
        item._update_parent_and_storage_key(self, key)
        self._data[key] = item

    #
    # Define the ICategorizedObjectContainer abstract methods
    #

    def child(self, key):
        """Get the child object associated with a given
        storage key for this container.

        Raises:
            KeyError: if the argument is not a storage key
                for any children of this container
        """
        return self[key]

    def children(self):
        """A generator over the children of this container."""
        return self._data.values()

    #
    # Define the MutableMapping abstract methods
    #

    def __setitem__(self, key, item):
        if item.ctype is self.ctype:
            if item._parent is None:
                if key in self._data:
                    logger.warning(
                        "Implicitly replacing the entry %s (type=%s) "
                        "with a new object (type=%s). This is usually "
                        "indicative of a modeling error. To avoid this "
                        "warning, delete the original object from the "
                        "container before assigning a new object."
                        % (
                            self[key].name,
                            self[key].__class__.__name__,
                            item.__class__.__name__,
                        )
                    )
                    self._data[key]._clear_parent_and_storage_key()
                self._fast_insert(key, item)
                return
            elif (key in self._data) and (self._data[key] is item):
                # a very special case that makes sense to handle
                # because the implied order should be: (1) delete
                # the object at the current index, (2) insert the
                # the new object. This performs both without any
                # actions, but it is an extremely rare case, so
                # it should go last.
                return
            # see note about allowing categorized objects to
            # live in more than one container
            raise ValueError(
                "Invalid assignment to %s type with name '%s' "
                "at key %s. A parent container has already been "
                "assigned to the object being inserted: %s"
                % (self.__class__.__name__, self.name, key, item.parent.name)
            )
        else:
            raise TypeError(
                "Invalid assignment to type %s with index %s. "
                "The object being inserted has the wrong "
                "category type: %s" % (self.__class__.__name__, key, item.ctype)
            )

    def __delitem__(self, key):
        self._data[key]._clear_parent_and_storage_key()
        del self._data[key]

    def __getitem__(self, key):
        return self._data[key]

    def __iter__(self):
        return self._data.__iter__()

    def __len__(self):
        return self._data.__len__()

    #
    # Override a few default implementations on MutableMapping
    #

    def __contains__(self, key):
        return self._data.__contains__(key)

    # We want to avoid generating Pyomo expressions due to
    # comparison of values, so we convert both objects to a
    # plain dictionary mapping key->(type(val), id(val)) and
    # compare that instead.
    def __eq__(self, other):
        if not isinstance(other, collections.abc.Mapping):
            return False
        return {key: (type(val), id(val)) for key, val in self.items()} == {
            key: (type(val), id(val)) for key, val in other.items()
        }

    def __ne__(self, other):
        return not (self == other)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import io

from pyomo.common.tee import TeeStream, capture_output
from pyomo.contrib.solver.common.base import Availability
from pyomo.contrib.solver.solvers.knitro.api import (
    KNITRO_AVAILABLE,
    get_version,
    knitro,
)


class Package:
    """Manages the global KNITRO license context and provides utility methods for license handling.

    This class handles license initialization, release, context creation, version reporting,
    and license availability checks for the KNITRO solver.
    """

    _license_context = None

    @staticmethod
    def initialize_license():
        """Initialize the global KNITRO license context if not already initialized.

        Returns:
            The KNITRO license context object.

        """
        if Package._license_context is None:
            Package._license_context = knitro.KN_checkout_license()
        return Package._license_context

    @staticmethod
    def release_license() -> None:
        """Release the global KNITRO license context if it exists."""
        if Package._license_context is not None:
            knitro.KN_release_license(Package._license_context)
            Package._license_context = None

    @staticmethod
    def create_context():
        """Create a new KNITRO context using the global license context.

        Returns:
            The new KNITRO context object.

        """
        lmc = Package.initialize_license()
        return knitro.KN_new_lm(lmc)

    @staticmethod
    def get_version() -> tuple[int, int, int] | None:
        """Get the version of the KNITRO solver as a tuple.

        Returns:
            tuple[int, int, int]: The (major, minor, patch) version of KNITRO
            or None if KNITRO version could not be determined.

        """
        version = get_version()
        if version is None:
            return None
        major, minor, patch = map(int, version.split("."))
        return major, minor, patch

    @staticmethod
    def check_availability() -> Availability:
        """Check if the KNITRO solver and license are available.

        Returns:
            Availability: The availability status (FullLicense, BadLicense, NotFound).

        """
        if not bool(KNITRO_AVAILABLE):
            return Availability.NotFound
        try:
            stream = io.StringIO()
            with capture_output(TeeStream(stream), capture_fd=True):
                kc = Package.create_context()
                knitro.KN_free(kc)
            # TODO: parse the stream to check the license type.
            return Availability.FullLicense
        except Exception:
            return Availability.BadLicense


class PackageChecker:
    _available_cache: Availability | None

    def __init__(self) -> None:
        self._available_cache = None

    def available(self) -> Availability:
        if self._available_cache is None:
            self._available_cache = Package.check_availability()
        return self._available_cache

    def version(self) -> tuple[int, int, int] | None:
        return Package.get_version()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.errors import IterationLimitError
from pyomo.common.numeric_types import native_numeric_types, native_complex_types, value
from pyomo.core.expr.calculus.derivatives import differentiate
from pyomo.core.base.constraint import Constraint

import logging

logger = logging.getLogger(__name__)

_default_differentiation_mode = differentiate.Modes.sympy
_symbolic_modes = {
    None,
    differentiate.Modes.sympy,
    differentiate.Modes.reverse_symbolic,
}


def calculate_variable_from_constraint(
    variable,
    constraint,
    eps=1e-8,
    iterlim=1000,
    linesearch=True,
    alpha_min=1e-8,
    diff_mode=None,
):
    """Calculate the variable value given a specified equality constraint

    This function calculates the value of the specified variable
    necessary to make the provided equality constraint feasible
    (assuming any other variables values are fixed).  The method first
    attempts to solve for the variable value assuming it appears
    linearly in the constraint.  If that doesn't converge the constraint
    residual, it falls back on Newton's method using exact (symbolic)
    derivatives.

    Notes
    -----
    This is an unconstrained solver and is NOT guaranteed to respect the
    variable bounds or domain.  The solver may leave the variable value
    in an infeasible state (outside the declared bounds or domain bounds).

    Parameters:
    -----------
    variable: :py:class:`VarData`
        The variable to solve for
    constraint: :py:class:`ConstraintData` or relational expression or `tuple`
        The equality constraint to use to solve for the variable value.
        May be a `ConstraintData` object or any valid argument for
        ``Constraint(expr=<>)`` (i.e., a relational expression or 2- or
        3-tuple)
    eps: `float`
        The tolerance to use to determine equality [default=1e-8].
    iterlim: `int`
        The maximum number of iterations if this method has to fall back
        on using Newton's method.  Raises RuntimeError on iteration
        limit [default=1000]
    linesearch: `bool`
        Decides whether or not to use the linesearch (recommended).
        [default=True]
    alpha_min: `float`
        The minimum fractional step to use in the linesearch [default=1e-8].
    diff_mode: :py:enum:`pyomo.core.expr.calculus.derivatives.Modes`
        The mode to use to differentiate the expression.  If
        unspecified, defaults to `Modes.sympy`

    Returns:
    --------
    None

    """
    # Leverage all the Constraint logic to process the incoming tuple/expression
    if not getattr(constraint, 'ctype', None) is Constraint:
        constraint = Constraint(expr=constraint, name=type(constraint).__name__)
        constraint.construct()

    if constraint.is_indexed():
        raise ValueError(
            'calculate_variable_from_constraint(): constraint must be a '
            'scalar constraint or a single ConstraintData.  Received '
            f'{constraint.__class__.__name__} ("{constraint.name}")'
        )

    body = constraint.body
    lower = constraint.lb
    upper = constraint.ub

    if lower != upper:
        raise ValueError(f"Constraint '{constraint}' must be an equality constraint")

    _invalid_types = set(native_complex_types)
    _invalid_types.add(type(None))

    if variable.value is None:
        # Note that we use "skip_validation=True" here as well, as the
        # variable domain may not admit the calculated initial guesses,
        # and we want to bypass that check.
        if variable.lb is None:
            if variable.ub is None:
                # no variable values, and no lower or upper bound - set
                # initial value to 0.0
                variable.set_value(0, skip_validation=True)
            else:
                # no variable value or lower bound - set to 0 or upper
                # bound whichever is lower
                variable.set_value(min(0, variable.ub), skip_validation=True)
        elif variable.ub is None:
            # no variable value or upper bound - set to 0 or lower
            # bound, whichever is higher
            variable.set_value(max(0, variable.lb), skip_validation=True)
        else:
            # we have upper and lower bounds
            if variable.lb <= 0 and variable.ub >= 0:
                # set the initial value to 0 if bounds bracket 0
                variable.set_value(0, skip_validation=True)
            else:
                # set the initial value to the midpoint of the bounds
                variable.set_value(
                    (variable.lb + variable.ub) / 2.0, skip_validation=True
                )

    # store the initial value to use later if necessary
    orig_initial_value = variable.value

    # solve the common case where variable is linear with coefficient of 1.0
    x1 = value(variable)
    # Note: both the direct (linear) calculation and Newton's method
    # below rely on a numerically valid initial starting point.
    # While we have strategies for dealing with hitting numerically
    # invalid (e.g., sqrt(-1)) conditions below, if the initial point is
    # not valid, we will allow that exception to propagate up
    try:
        residual_1 = value(body)
    except:
        logger.error(
            "Encountered an error evaluating the expression at the "
            "initial guess.\n\tPlease provide a different initial guess."
        )
        raise

    try:
        variable.set_value(x1 - (residual_1 - upper), skip_validation=True)
        residual_2 = value(body, exception=False)
    except OverflowError:
        # If we encounter an error while evaluating the expression at the
        # linear intercept calculated assuming the derivative was 1.  This
        # is most commonly due to nonlinear expressions (like sqrt())
        # becoming invalid/complex.  We will skip the rest of the
        # "shortcuts" that assume the expression is linear and move directly
        # to using Newton's method.
        residual_2 = None

    if residual_2.__class__ not in _invalid_types:
        # if the variable appears linearly with a coefficient of 1, then we
        # are done
        if abs(residual_2 - upper) < eps:
            # Re-set the variable value to trigger any warnings WRT the
            # final variable state
            variable.set_value(variable.value)
            return

        # Assume the variable appears linearly and calculate the coefficient
        x2 = value(variable)
        slope = float(residual_1 - residual_2) / (x1 - x2)
        intercept = (residual_1 - upper) - slope * x1
        if slope:
            variable.set_value(-intercept / slope, skip_validation=True)
            try:
                body_val = value(body, exception=False)
            except OverflowError:
                body_val = None
            if body_val.__class__ not in _invalid_types and abs(body_val - upper) < eps:
                # Re-set the variable value to trigger any warnings WRT
                # the final variable state
                variable.set_value(variable.value)
                return

    # Variable appears nonlinearly; solve using Newton's method
    #
    # restore initial value
    variable.set_value(orig_initial_value, skip_validation=True)
    expr = body - upper

    expr_deriv = None
    if diff_mode in _symbolic_modes:
        try:
            expr_deriv = differentiate(
                expr, wrt=variable, mode=diff_mode or _default_differentiation_mode
            )
        except:
            if diff_mode is None:
                # If the user didn't care how we differentiate, try to
                # (mostly silently) revert to numeric differentiation.
                logger.debug(
                    'Calculating symbolic derivative of expression failed. '
                    'Reverting to numeric differentiation'
                )
                diff_mode = differentiate.Modes.reverse_numeric
            else:
                raise

        if type(expr_deriv) in native_numeric_types and expr_deriv == 0:
            raise ValueError(
                f"Variable '{variable}' derivative == 0 in constraint "
                f"'{constraint}', cannot solve for variable"
            )

    if expr_deriv is None:
        fp0 = differentiate(expr, wrt=variable, mode=diff_mode)
    else:
        fp0 = value(expr_deriv)

    if abs(value(fp0)) < 1e-12:
        raise ValueError(
            f"Initial value for variable '{variable}' results in a derivative "
            f"value for constraint '{constraint}' that is very close to zero.\n"
            "\tPlease provide a different initial guess."
        )

    iter_left = iterlim
    fk = residual_1 - upper
    while abs(fk) > eps and iter_left:
        iter_left -= 1
        if not iter_left:
            raise IterationLimitError(
                f"Iteration limit (%s) reached solving for variable '{variable}' "
                f"using constraint '{constraint}'; remaining residual = %s"
                % (iterlim, value(expr))
            )

        # compute step
        xk = value(variable)
        try:
            fk = value(expr)
            if fk.__class__ in _invalid_types and fk is not None:
                raise ValueError("Complex numbers are not allowed in Newton's method.")
        except:
            # We hit numerical problems with the last step (possible if
            # the line search is turned off)
            logger.error(
                "Newton's method encountered an error evaluating the "
                f"expression for constraint '{constraint}'.\n\tPlease provide a "
                "different initial guess or enable the linesearch if you have not."
            )
            raise

        if expr_deriv is None:
            fpk = differentiate(expr, wrt=variable, mode=diff_mode)
        else:
            fpk = value(expr_deriv)

        if abs(fpk) < 1e-12:
            # TODO: should this raise a ValueError or a new
            # DerivativeError (subclassing ArithmeticError)?
            raise RuntimeError(
                "Newton's method encountered a derivative of constraint "
                f"'{constraint}' with respect to variable '{variable}' that was too "
                "close to zero.\n\tPlease provide a different initial guess "
                "or enable the linesearch if you have not."
            )
        pk = -fk / fpk
        alpha = 1.0
        xkp1 = xk + alpha * pk
        variable.set_value(xkp1, skip_validation=True)

        # perform line search
        if linesearch:
            c1 = 0.999  # ensure sufficient progress
            while alpha > alpha_min:
                # check if the value at xkp1 has sufficient reduction in
                # the residual
                try:
                    fkp1 = value(expr, exception=False)
                    # HACK for Python3 support, pending resolution of #879
                    # Issue #879 also pertains to other checks for "complex"
                    # in this method.
                    if fkp1.__class__ in _invalid_types:
                        # We cannot perform computations on complex numbers
                        fkp1 = None
                    if fkp1 is not None and fkp1**2 < c1 * fk**2:
                        # found an alpha value with sufficient reduction
                        # continue to the next step
                        fk = fkp1
                        break
                except OverflowError:
                    # Encountered an overflow, either from evaluating
                    # this point in the line search (to get fkp1) or
                    # from squaring fkp1.  (The example from #3540
                    # actually triggers both).  Reject this alpha value.
                    pass
                alpha /= 2.0
                xkp1 = xk + alpha * pk
                variable.set_value(xkp1, skip_validation=True)

            if alpha <= alpha_min:
                residual = value(expr, exception=False)
                if residual.__class__ in _invalid_types:
                    residual = "{function evaluation error}"
                raise IterationLimitError(
                    f"Linesearch iteration limit reached solving for "
                    f"variable '{variable}' using constraint '{constraint}'; "
                    f"remaining residual = {residual}."
                )
    #
    # Re-set the variable value to trigger any warnings WRT the final
    # variable state
    variable.set_value(variable.value)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.config import Bool, ConfigValue
from pyomo.contrib.solver.common.config import SolverConfig


class KnitroConfig(SolverConfig):
    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ) -> None:
        super().__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )

        self.rebuild_model_on_remove_var: bool = self.declare(
            "rebuild_model_on_remove_var",
            ConfigValue(
                domain=Bool,
                default=False,
                doc=(
                    "KNITRO solver does not allow variable removal. We can "
                    "either make the variable a continuous free variable or "
                    "rebuild the whole model when variable removal is "
                    "attempted. When `rebuild_model_on_remove_var` is set to "
                    "True, the model will be rebuilt."
                ),
            ),
        )

        self.restore_variable_values_after_solve: bool = self.declare(
            "restore_variable_values_after_solve",
            ConfigValue(
                domain=Bool,
                default=False,
                doc=(
                    "To evaluate non-linear constraints, KNITRO solver sets "
                    "explicit values on variables. This option controls "
                    "whether to restore the original variable values after "
                    "solving."
                ),
            ),
        )



# -*- coding: utf-8 -*-
# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
#
import math
import pickle

from pyomo.common.errors import PyomoException
import pyomo.common.unittest as unittest
from pyomo.environ import (
    ConcreteModel,
    Var,
    Param,
    Set,
    Constraint,
    Objective,
    Expression,
    ExternalFunction,
    value,
    sum_product,
    maximize,
    units,
    log,
    log10,
    exp,
    sqrt,
    cos,
    sin,
    tan,
    asin,
    acos,
    atan,
    cosh,
    sinh,
    tanh,
    asinh,
    acosh,
    atanh,
    ceil,
    floor,
    BooleanVar,
)
from pyomo.common.log import LoggingIntercept
from pyomo.util.check_units import assert_units_consistent, check_units_equivalent
from pyomo.core.expr import inequality
from pyomo.core.expr.numvalue import NumericConstant
import pyomo.core.expr as EXPR
from pyomo.core.base.units_container import (
    pint_available,
    pint_module,
    _DeferredUnitsSingleton,
    InconsistentUnitsError,
    UnitsError,
    PyomoUnitsContainer,
    as_quantity,
    _PyomoUnit,
)
from io import StringIO


def python_callback_function(arg1, arg2):
    return 42.0


@unittest.skipIf(not pint_available, 'Testing units requires pint')
class TestPyomoUnit(unittest.TestCase):
    def test_container_constructor(self):
        # Custom pint registry:
        um0 = PyomoUnitsContainer(None)
        self.assertIsNone(um0.pint_registry)
        self.assertIsNone(um0._pint_dimensionless)
        um1 = PyomoUnitsContainer()
        self.assertIsNotNone(um1.pint_registry)
        self.assertIsNotNone(um1._pint_dimensionless)
        with self.assertRaisesRegex(
            ValueError, 'Cannot operate with Unit and Unit of different registries'
        ):
            self.assertEqual(um1._pint_dimensionless, units._pint_dimensionless)
        self.assertIsNot(um1.pint_registry, units.pint_registry)
        um2 = PyomoUnitsContainer(pint_module.UnitRegistry())
        self.assertIsNotNone(um2.pint_registry)
        self.assertIsNotNone(um2._pint_dimensionless)
        with self.assertRaisesRegex(
            ValueError, 'Cannot operate with Unit and Unit of different registries'
        ):
            self.assertEqual(um2._pint_dimensionless, units._pint_dimensionless)
        self.assertIsNot(um2.pint_registry, units.pint_registry)
        self.assertIsNot(um2.pint_registry, um1.pint_registry)

        um3 = PyomoUnitsContainer(units.pint_registry)
        self.assertIs(um3.pint_registry, units.pint_registry)
        self.assertEqual(um3._pint_dimensionless, units._pint_dimensionless)

    def test_PyomoUnit_NumericValueMethods(self):
        m = ConcreteModel()
        uc = units
        kg = uc.kg

        self.assertEqual(kg.getname(), 'kg')
        self.assertEqual(kg.name, 'kg')
        self.assertEqual(kg.local_name, 'kg')

        m.kg = uc.kg

        self.assertEqual(m.kg.name, 'kg')
        self.assertEqual(m.kg.local_name, 'kg')

        self.assertEqual(kg.is_constant(), False)
        self.assertEqual(kg.is_fixed(), True)
        self.assertEqual(kg.is_parameter_type(), False)
        self.assertEqual(kg.is_variable_type(), False)
        self.assertEqual(kg.is_potentially_variable(), False)
        self.assertEqual(kg.is_named_expression_type(), False)
        self.assertEqual(kg.is_expression_type(), False)
        self.assertEqual(kg.is_component_type(), False)
        self.assertEqual(kg.is_expression_type(EXPR.ExpressionType.RELATIONAL), False)
        self.assertEqual(kg.is_indexed(), False)
        self.assertEqual(kg._compute_polynomial_degree(None), 0)

        with self.assertRaises(TypeError):
            x = float(kg)
        with self.assertRaises(TypeError):
            x = int(kg)

        assert_units_consistent(kg < m.kg)
        assert_units_consistent(kg > m.kg)
        assert_units_consistent(kg <= m.kg)
        assert_units_consistent(kg >= m.kg)
        assert_units_consistent(kg == m.kg)
        assert_units_consistent(kg + m.kg)
        assert_units_consistent(kg - m.kg)

        with self.assertRaises(InconsistentUnitsError):
            assert_units_consistent(kg + 3)

        with self.assertRaises(InconsistentUnitsError):
            assert_units_consistent(kg - 3)

        with self.assertRaises(InconsistentUnitsError):
            assert_units_consistent(3 + kg)

        with self.assertRaises(InconsistentUnitsError):
            assert_units_consistent(3 - kg)

        # should not assert
        # check __mul__
        self.assertEqual(str(uc.get_units(kg * 3)), 'kg')
        # check __rmul__
        self.assertEqual(str(uc.get_units(3 * kg)), 'kg')
        # check div / truediv
        self.assertEqual(str(uc.get_units(kg / 3.0)), 'kg')
        # check rdiv / rtruediv
        self.assertEqual(str(uc.get_units(3.0 / kg)), '1/kg')
        # check pow
        self.assertEqual(str(uc.get_units(kg**2)), 'kg**2')

        # check rpow
        x = 2**kg  # creation is allowed, only fails when units are "checked"
        with self.assertRaises(UnitsError):
            assert_units_consistent(x)

        x = kg
        x += kg
        self.assertEqual(str(uc.get_units(x)), 'kg')

        x = kg
        x -= 2.0 * kg
        self.assertEqual(str(uc.get_units(x)), 'kg')

        x = kg
        x *= 3
        self.assertEqual(str(uc.get_units(x)), 'kg')

        x = kg
        x **= 3
        self.assertEqual(str(uc.get_units(x)), 'kg**3')

        self.assertEqual(str(uc.get_units(-kg)), 'kg')
        self.assertEqual(str(uc.get_units(+kg)), 'kg')
        self.assertEqual(str(uc.get_units(abs(kg))), 'kg')

        self.assertEqual(str(kg), 'kg')
        self.assertEqual(kg.to_string(), 'kg')
        # ToDo: is this really the correct behavior for verbose?
        self.assertEqual(kg.to_string(verbose=True), 'kg')
        self.assertEqual((kg / uc.s).to_string(), 'kg/s')
        self.assertEqual((kg * uc.m**2 / uc.s).to_string(), 'kg*m**2/s')

        m.v = Var(initialize=3, units=uc.J)
        e = uc.convert(m.v, uc.g * uc.m**2 / uc.s**2)
        self.assertEqual(e.to_string(), '1000.0*(g*m**2/J/s**2)*v')

        # check __nonzero__ / __bool__
        with self.assertRaisesRegex(
            PyomoException,
            r"Cannot convert non-constant Pyomo " r"numeric value \(kg\) to bool.",
        ):
            bool(kg)

        # __call__ returns 1.0
        self.assertEqual(kg(), 1.0)
        self.assertEqual(value(kg), 1.0)

        # test pprint
        buf = StringIO()
        kg.pprint(ostream=buf)
        self.assertEqual('kg', buf.getvalue())

        # test str representations for dimensionless
        dless = uc.dimensionless
        self.assertEqual('dimensionless', str(dless))

    def _get_check_units_ok(
        self, x, pyomo_units_container, str_check=None, expected_type=None
    ):
        if expected_type is not None:
            self.assertEqual(expected_type, type(x))

        assert_units_consistent(x)
        if str_check is not None:
            self.assertEqual(str_check, str(pyomo_units_container.get_units(x)))
        else:
            # if str_check is None, then we expect the units to be None
            self.assertIsNone(pyomo_units_container.get_units(x))

    def _get_check_units_fail(
        self,
        x,
        pyomo_units_container,
        expected_type=None,
        expected_error=InconsistentUnitsError,
    ):
        if expected_type is not None:
            self.assertEqual(expected_type, type(x))

        with self.assertRaises(expected_error):
            assert_units_consistent(x)

        # we also expect get_units to fail
        with self.assertRaises(expected_error):
            pyomo_units_container.get_units(x)

    def test_get_check_units_on_all_expressions(self):
        # this method is going to test all the expression types that should work
        # to be defensive, we will also test that we actually have the expected expression type
        # therefore, if the expression system changes and we get a different expression type,
        # we will know we need to change these tests

        uc = units
        kg = uc.kg
        m = uc.m

        model = ConcreteModel()
        model.x = Var()
        model.y = Var()
        model.z = Var()
        model.p = Param(initialize=42.0, mutable=True)
        model.xkg = Var(units=kg)
        model.ym = Var(units=m)

        # test equality
        self._get_check_units_ok(
            3.0 * kg == 1.0 * kg, uc, 'kg', EXPR.EqualityExpression
        )
        self._get_check_units_fail(3.0 * kg == 2.0 * m, uc, EXPR.EqualityExpression)

        # test inequality
        self._get_check_units_ok(
            3.0 * kg <= 1.0 * kg, uc, 'kg', EXPR.InequalityExpression
        )
        self._get_check_units_fail(3.0 * kg <= 2.0 * m, uc, EXPR.InequalityExpression)
        self._get_check_units_ok(
            3.0 * kg >= 1.0 * kg, uc, 'kg', EXPR.InequalityExpression
        )
        self._get_check_units_fail(3.0 * kg >= 2.0 * m, uc, EXPR.InequalityExpression)

        # test RangedExpression
        self._get_check_units_ok(
            inequality(3.0 * kg, 4.0 * kg, 5.0 * kg), uc, 'kg', EXPR.RangedExpression
        )
        self._get_check_units_fail(
            inequality(3.0 * m, 4.0 * kg, 5.0 * kg), uc, EXPR.RangedExpression
        )
        self._get_check_units_fail(
            inequality(3.0 * kg, 4.0 * m, 5.0 * kg), uc, EXPR.RangedExpression
        )
        self._get_check_units_fail(
            inequality(3.0 * kg, 4.0 * kg, 5.0 * m), uc, EXPR.RangedExpression
        )

        # test SumExpression, NPV_SumExpression
        self._get_check_units_ok(
            3.0 * model.x * kg + 1.0 * model.y * kg + 3.65 * model.z * kg,
            uc,
            'kg',
            EXPR.LinearExpression,
        )
        self._get_check_units_fail(
            3.0 * model.x * kg + 1.0 * model.y * m + 3.65 * model.z * kg,
            uc,
            EXPR.LinearExpression,
        )

        self._get_check_units_ok(
            3.0 * kg + 1.0 * kg + 2.0 * kg, uc, 'kg', EXPR.NPV_SumExpression
        )
        self._get_check_units_fail(
            3.0 * kg + 1.0 * kg + 2.0 * m, uc, EXPR.NPV_SumExpression
        )

        # test ProductExpression, NPV_ProductExpression
        self._get_check_units_ok(
            model.x * kg * model.y * m, uc, 'kg*m', EXPR.ProductExpression
        )
        self._get_check_units_ok(
            3.0 * kg * 1.0 * m, uc, 'kg*m', EXPR.NPV_ProductExpression
        )
        self._get_check_units_ok(3.0 * kg * m, uc, 'kg*m', EXPR.NPV_ProductExpression)
        # I don't think that there are combinations that can "fail" for products

        # test MonomialTermExpression
        self._get_check_units_ok(model.x * kg, uc, 'kg', EXPR.MonomialTermExpression)

        # test DivisionExpression, NPV_DivisionExpression
        self._get_check_units_ok(
            1.0 / (model.x * kg), uc, '1/kg', EXPR.DivisionExpression
        )
        self._get_check_units_ok(2.0 / kg, uc, '1/kg', EXPR.NPV_DivisionExpression)
        self._get_check_units_ok(
            (model.x * kg) / 1.0, uc, 'kg', EXPR.MonomialTermExpression
        )
        self._get_check_units_ok(kg / 2.0, uc, 'kg', EXPR.NPV_DivisionExpression)
        self._get_check_units_ok(
            model.y * m / (model.x * kg), uc, 'm/kg', EXPR.DivisionExpression
        )
        self._get_check_units_ok(m / kg, uc, 'm/kg', EXPR.NPV_DivisionExpression)
        # I don't think that there are combinations that can "fail" for products

        # test PowExpression, NPV_PowExpression
        self._get_check_units_ok(kg**2, uc, 'kg**2', EXPR.NPV_PowExpression)
        self._get_check_units_ok(
            model.p**model.p, uc, 'dimensionless', EXPR.NPV_PowExpression
        )
        self._get_check_units_ok(kg**model.p, uc, 'kg**42', EXPR.NPV_PowExpression)
        self._get_check_units_ok(
            (model.x * kg**2) ** 3, uc, 'kg**6', EXPR.PowExpression
        )
        self._get_check_units_fail(kg**model.x, uc, EXPR.PowExpression, UnitsError)
        self._get_check_units_fail(model.x**kg, uc, EXPR.PowExpression, UnitsError)
        self._get_check_units_ok(kg**2, uc, 'kg**2', EXPR.NPV_PowExpression)
        self._get_check_units_fail(3.0**kg, uc, EXPR.NPV_PowExpression, UnitsError)

        # test NegationExpression, NPV_NegationExpression
        self._get_check_units_ok(
            -(kg * model.x * model.y), uc, 'kg', EXPR.NegationExpression
        )
        self._get_check_units_ok(-kg, uc, 'kg', EXPR.NPV_NegationExpression)
        # don't think there are combinations that fan "fail" for negation

        # test AbsExpression, NPV_AbsExpression
        self._get_check_units_ok(abs(kg * model.x), uc, 'kg', EXPR.AbsExpression)
        self._get_check_units_ok(abs(kg), uc, 'kg', EXPR.NPV_AbsExpression)
        # don't think there are combinations that fan "fail" for abs

        # test the different UnaryFunctionExpression / NPV_UnaryFunctionExpression types
        # log
        self._get_check_units_ok(
            log(3.0 * model.x), uc, 'dimensionless', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_fail(
            log(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_ok(
            log(3.0 * model.p), uc, 'dimensionless', EXPR.NPV_UnaryFunctionExpression
        )
        self._get_check_units_fail(
            log(3.0 * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # log10
        self._get_check_units_ok(
            log10(3.0 * model.x), uc, 'dimensionless', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_fail(
            log10(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_ok(
            log10(3.0 * model.p), uc, 'dimensionless', EXPR.NPV_UnaryFunctionExpression
        )
        self._get_check_units_fail(
            log10(3.0 * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # sin
        self._get_check_units_ok(
            sin(3.0 * model.x), uc, 'dimensionless', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_ok(
            sin(3.0 * model.x * uc.radians),
            uc,
            'dimensionless',
            EXPR.UnaryFunctionExpression,
        )
        self._get_check_units_fail(
            sin(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_fail(
            sin(3.0 * kg * model.x * uc.kg),
            uc,
            EXPR.UnaryFunctionExpression,
            UnitsError,
        )
        self._get_check_units_ok(
            sin(3.0 * model.p * uc.radians),
            uc,
            'dimensionless',
            EXPR.NPV_UnaryFunctionExpression,
        )
        self._get_check_units_fail(
            sin(3.0 * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # cos
        self._get_check_units_ok(
            cos(3.0 * model.x * uc.radians),
            uc,
            'dimensionless',
            EXPR.UnaryFunctionExpression,
        )
        self._get_check_units_fail(
            cos(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_fail(
            cos(3.0 * kg * model.x * uc.kg),
            uc,
            EXPR.UnaryFunctionExpression,
            UnitsError,
        )
        self._get_check_units_ok(
            cos(3.0 * model.p * uc.radians),
            uc,
            'dimensionless',
            EXPR.NPV_UnaryFunctionExpression,
        )
        self._get_check_units_fail(
            cos(3.0 * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # tan
        self._get_check_units_ok(
            tan(3.0 * model.x * uc.radians),
            uc,
            'dimensionless',
            EXPR.UnaryFunctionExpression,
        )
        self._get_check_units_fail(
            tan(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_fail(
            tan(3.0 * kg * model.x * uc.kg),
            uc,
            EXPR.UnaryFunctionExpression,
            UnitsError,
        )
        self._get_check_units_ok(
            tan(3.0 * model.p * uc.radians),
            uc,
            'dimensionless',
            EXPR.NPV_UnaryFunctionExpression,
        )
        self._get_check_units_fail(
            tan(3.0 * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # sin
        self._get_check_units_ok(
            sinh(3.0 * model.x * uc.radians),
            uc,
            'dimensionless',
            EXPR.UnaryFunctionExpression,
        )
        self._get_check_units_fail(
            sinh(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_fail(
            sinh(3.0 * kg * model.x * uc.kg),
            uc,
            EXPR.UnaryFunctionExpression,
            UnitsError,
        )
        self._get_check_units_ok(
            sinh(3.0 * model.p * uc.radians),
            uc,
            'dimensionless',
            EXPR.NPV_UnaryFunctionExpression,
        )
        self._get_check_units_fail(
            sinh(3.0 * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # cos
        self._get_check_units_ok(
            cosh(3.0 * model.x * uc.radians),
            uc,
            'dimensionless',
            EXPR.UnaryFunctionExpression,
        )
        self._get_check_units_fail(
            cosh(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_fail(
            cosh(3.0 * kg * model.x * uc.kg),
            uc,
            EXPR.UnaryFunctionExpression,
            UnitsError,
        )
        self._get_check_units_ok(
            cosh(3.0 * model.p * uc.radians),
            uc,
            'dimensionless',
            EXPR.NPV_UnaryFunctionExpression,
        )
        self._get_check_units_fail(
            cosh(3.0 * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # tan
        self._get_check_units_ok(
            tanh(3.0 * model.x * uc.radians),
            uc,
            'dimensionless',
            EXPR.UnaryFunctionExpression,
        )
        self._get_check_units_fail(
            tanh(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_fail(
            tanh(3.0 * kg * model.x * uc.kg),
            uc,
            EXPR.UnaryFunctionExpression,
            UnitsError,
        )
        self._get_check_units_ok(
            tanh(3.0 * model.p * uc.radians),
            uc,
            'dimensionless',
            EXPR.NPV_UnaryFunctionExpression,
        )
        self._get_check_units_fail(
            tanh(3.0 * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # asin
        self._get_check_units_ok(
            asin(3.0 * model.x), uc, 'rad', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_fail(
            asin(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_ok(
            asin(3.0 * model.p), uc, 'rad', EXPR.NPV_UnaryFunctionExpression
        )
        self._get_check_units_fail(
            asin(3.0 * model.p * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # acos
        self._get_check_units_ok(
            acos(3.0 * model.x), uc, 'rad', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_fail(
            acos(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_ok(
            acos(3.0 * model.p), uc, 'rad', EXPR.NPV_UnaryFunctionExpression
        )
        self._get_check_units_fail(
            acos(3.0 * model.p * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # atan
        self._get_check_units_ok(
            atan(3.0 * model.x), uc, 'rad', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_fail(
            atan(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_ok(
            atan(3.0 * model.p), uc, 'rad', EXPR.NPV_UnaryFunctionExpression
        )
        self._get_check_units_fail(
            atan(3.0 * model.p * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # exp
        self._get_check_units_ok(
            exp(3.0 * model.x), uc, 'dimensionless', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_fail(
            exp(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_ok(
            exp(3.0 * model.p), uc, 'dimensionless', EXPR.NPV_UnaryFunctionExpression
        )
        self._get_check_units_fail(
            exp(3.0 * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # sqrt
        self._get_check_units_ok(
            sqrt(3.0 * model.x), uc, 'dimensionless', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_ok(
            sqrt(3.0 * model.x * kg**2), uc, 'kg', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_ok(
            sqrt(3.0 * model.x * kg), uc, 'kg**0.5', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_ok(
            sqrt(3.0 * model.p), uc, 'dimensionless', EXPR.NPV_UnaryFunctionExpression
        )
        self._get_check_units_ok(
            sqrt(3.0 * model.p * kg**2), uc, 'kg', EXPR.NPV_UnaryFunctionExpression
        )
        self._get_check_units_ok(
            sqrt(3.0 * model.p * kg), uc, 'kg**0.5', EXPR.NPV_UnaryFunctionExpression
        )
        # asinh
        self._get_check_units_ok(
            asinh(3.0 * model.x), uc, 'rad', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_fail(
            asinh(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_ok(
            asinh(3.0 * model.p), uc, 'rad', EXPR.NPV_UnaryFunctionExpression
        )
        self._get_check_units_fail(
            asinh(3.0 * model.p * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # acosh
        self._get_check_units_ok(
            acosh(3.0 * model.x), uc, 'rad', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_fail(
            acosh(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_ok(
            acosh(3.0 * model.p), uc, 'rad', EXPR.NPV_UnaryFunctionExpression
        )
        self._get_check_units_fail(
            acosh(3.0 * model.p * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # atanh
        self._get_check_units_ok(
            atanh(3.0 * model.x), uc, 'rad', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_fail(
            atanh(3.0 * kg * model.x), uc, EXPR.UnaryFunctionExpression, UnitsError
        )
        self._get_check_units_ok(
            atanh(3.0 * model.p), uc, 'rad', EXPR.NPV_UnaryFunctionExpression
        )
        self._get_check_units_fail(
            atanh(3.0 * model.p * kg), uc, EXPR.NPV_UnaryFunctionExpression, UnitsError
        )
        # ceil
        self._get_check_units_ok(
            ceil(kg * model.x), uc, 'kg', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_ok(ceil(kg), uc, 'kg', EXPR.NPV_UnaryFunctionExpression)
        # don't think there are combinations that fan "fail" for ceil
        # floor
        self._get_check_units_ok(
            floor(kg * model.x), uc, 'kg', EXPR.UnaryFunctionExpression
        )
        self._get_check_units_ok(floor(kg), uc, 'kg', EXPR.NPV_UnaryFunctionExpression)
        # don't think there are combinations that fan "fail" for floor

        # test Expr_ifExpression
        # consistent if, consistent then/else
        self._get_check_units_ok(
            EXPR.Expr_if(
                IF=model.x * kg + kg >= 2.0 * kg, THEN=model.x * kg, ELSE=model.y * kg
            ),
            uc,
            'kg',
            EXPR.Expr_ifExpression,
        )
        # unitless if, consistent then/else
        self._get_check_units_ok(
            EXPR.Expr_if(IF=model.x >= 2.0, THEN=model.x * kg, ELSE=model.y * kg),
            uc,
            'kg',
            EXPR.Expr_ifExpression,
        )
        # consistent if, unitless then/else
        self._get_check_units_ok(
            EXPR.Expr_if(IF=model.x * kg + kg >= 2.0 * kg, THEN=model.x, ELSE=model.x),
            uc,
            'dimensionless',
            EXPR.Expr_ifExpression,
        )
        # inconsistent then/else
        self._get_check_units_fail(
            EXPR.Expr_if(IF=model.x >= 2.0, THEN=model.x * m, ELSE=model.y * kg),
            uc,
            EXPR.Expr_ifExpression,
        )
        # inconsistent then/else NPV
        self._get_check_units_fail(
            EXPR.Expr_if(IF=model.x >= 2.0, THEN=model.p * m, ELSE=model.p * kg),
            uc,
            EXPR.Expr_ifExpression,
        )
        # inconsistent then/else NPV units only
        self._get_check_units_fail(
            EXPR.Expr_if(IF=model.x >= 2.0, THEN=m, ELSE=kg), uc, EXPR.Expr_ifExpression
        )

        # test EXPR.IndexTemplate and GetItemExpression
        model.S = Set()
        i = EXPR.IndexTemplate(model.S)
        j = EXPR.IndexTemplate(model.S)
        self._get_check_units_ok(i, uc, 'dimensionless', EXPR.IndexTemplate)

        model.mat = Var(model.S, model.S)
        self._get_check_units_ok(
            model.mat[i, j + 1], uc, 'dimensionless', EXPR.Numeric_GetItemExpression
        )

        # test ExternalFunctionExpression, NPV_ExternalFunctionExpression
        model.ef = ExternalFunction(python_callback_function)
        self._get_check_units_ok(
            model.ef(model.x, model.y),
            uc,
            'dimensionless',
            EXPR.ExternalFunctionExpression,
        )
        self._get_check_units_ok(
            model.ef(1.0, 2.0), uc, 'dimensionless', EXPR.NPV_ExternalFunctionExpression
        )
        self._get_check_units_fail(
            model.ef(model.x * kg, model.y),
            uc,
            EXPR.ExternalFunctionExpression,
            UnitsError,
        )
        self._get_check_units_fail(
            model.ef(2.0 * kg, 1.0), uc, EXPR.NPV_ExternalFunctionExpression, UnitsError
        )

        # test ExternalFunctionExpression, NPV_ExternalFunctionExpression
        model.ef2 = ExternalFunction(python_callback_function, units=uc.kg)
        self._get_check_units_ok(
            model.ef2(model.x, model.y), uc, 'kg', EXPR.ExternalFunctionExpression
        )
        self._get_check_units_ok(
            model.ef2(1.0, 2.0), uc, 'kg', EXPR.NPV_ExternalFunctionExpression
        )
        self._get_check_units_fail(
            model.ef2(model.x * kg, model.y),
            uc,
            EXPR.ExternalFunctionExpression,
            UnitsError,
        )
        self._get_check_units_fail(
            model.ef2(2.0 * kg, 1.0),
            uc,
            EXPR.NPV_ExternalFunctionExpression,
            UnitsError,
        )

        # test ExternalFunctionExpression, NPV_ExternalFunctionExpression
        model.ef3 = ExternalFunction(
            python_callback_function, units=uc.kg, arg_units=[uc.kg, uc.m]
        )
        self._get_check_units_fail(
            model.ef3(model.x, model.y), uc, EXPR.ExternalFunctionExpression
        )
        self._get_check_units_fail(
            model.ef3(1.0, 2.0), uc, EXPR.NPV_ExternalFunctionExpression
        )
        self._get_check_units_fail(
            model.ef3(model.x * kg, model.y),
            uc,
            EXPR.ExternalFunctionExpression,
            UnitsError,
        )
        self._get_check_units_fail(
            model.ef3(2.0 * kg, 1.0),
            uc,
            EXPR.NPV_ExternalFunctionExpression,
            UnitsError,
        )
        self._get_check_units_ok(
            model.ef3(2.0 * kg, 1.0 * uc.m),
            uc,
            'kg',
            EXPR.NPV_ExternalFunctionExpression,
        )
        self._get_check_units_ok(
            model.ef3(model.x * kg, model.y * m),
            uc,
            'kg',
            EXPR.ExternalFunctionExpression,
        )
        self._get_check_units_ok(
            model.ef3(model.xkg, model.ym), uc, 'kg', EXPR.ExternalFunctionExpression
        )
        self._get_check_units_fail(
            model.ef3(model.ym, model.xkg),
            uc,
            EXPR.ExternalFunctionExpression,
            InconsistentUnitsError,
        )

    # @unittest.skip('Skipped testing LinearExpression since StreamBasedExpressionVisitor does not handle LinearExpressions')
    def test_linear_expression(self):
        uc = units
        model = ConcreteModel()
        kg = uc.kg
        m = uc.m

        # test LinearExpression
        # ToDo: Once this test is working correctly, this code should be moved to the test above
        model.vv = Var(['A', 'B', 'C'])
        self._get_check_units_ok(
            sum_product(model.vv), uc, 'dimensionless', EXPR.LinearExpression
        )

        linex1 = sum_product(
            model.vv, {'A': kg, 'B': kg, 'C': kg}, index=['A', 'B', 'C']
        )
        self._get_check_units_ok(linex1, uc, 'kg', EXPR.LinearExpression)

        linex2 = sum_product(
            model.vv, {'A': kg, 'B': m, 'C': kg}, index=['A', 'B', 'C']
        )
        self._get_check_units_fail(linex2, uc, EXPR.LinearExpression)

    def test_bad_units(self):
        uc = units
        with self.assertRaisesRegex(
            AttributeError, "Attribute unknown_bogus_unit not found."
        ):
            uc.unknown_bogus_unit

    def test_named_expression(self):
        uc = units
        m = ConcreteModel()
        m.x = Var(units=uc.kg)
        m.y = Var(units=uc.m)
        m.e = Expression(expr=m.x / m.y)
        self.assertEqual(str(uc.get_units(m.e)), 'kg/m')

    def test_dimensionless(self):
        uc = units
        kg = uc.kg
        dless = uc.dimensionless
        self._get_check_units_ok(
            2.0 == 2.0 * dless, uc, 'dimensionless', EXPR.EqualityExpression
        )
        x = uc.get_units(2.0)
        self.assertIs(type(x), _PyomoUnit)
        self.assertEqual(x, dless)
        x = uc.get_units(2.0 * dless)
        self.assertIs(type(x), _PyomoUnit)
        self.assertEqual(x, dless)
        x = uc.get_units(kg / kg)
        self.assertIs(type(x), _PyomoUnit)
        self.assertEqual(x, dless)

    def test_temperatures(self):
        uc = units

        # Pyomo units framework disallows "offset" units
        with self.assertRaises(UnitsError):
            degC = uc.celsius
        with self.assertRaises(UnitsError):
            degF = uc.degF

        # although we test delta versions here, users should not use these and
        # use absolute units instead
        delta_degC = uc.delta_degC
        K = uc.kelvin
        delta_degF = uc.delta_degF
        R = uc.rankine

        # In some recent versions of pint, rankine can be either
        # 'rankine' or '°R' (note UTF-8 encoding, which requires the
        # "coding: utf-8" comment flag at the top of this file).
        R_str = R.getname()
        # self.assertIn(R_str, ['rankine', '°R'])

        self._get_check_units_ok(2.0 * R + 3.0 * R, uc, R_str, EXPR.NPV_SumExpression)
        self._get_check_units_ok(2.0 * K + 3.0 * K, uc, 'K', EXPR.NPV_SumExpression)

        ex = 2.0 * delta_degC + 3.0 * delta_degC + 1.0 * delta_degC
        self.assertEqual(type(ex), EXPR.NPV_SumExpression)
        assert_units_consistent(ex)

        ex = 2.0 * delta_degF + 3.0 * delta_degF
        self.assertEqual(type(ex), EXPR.NPV_SumExpression)
        assert_units_consistent(ex)

        self._get_check_units_fail(2.0 * K + 3.0 * R, uc, EXPR.NPV_SumExpression)
        self._get_check_units_fail(
            2.0 * delta_degC + 3.0 * delta_degF, uc, EXPR.NPV_SumExpression
        )

        self.assertAlmostEqual(uc.convert_temp_K_to_C(323.15), 50.0, places=5)
        self.assertAlmostEqual(uc.convert_temp_C_to_K(50.0), 323.15, places=5)
        self.assertAlmostEqual(uc.convert_temp_R_to_F(509.67), 50.0, places=5)
        self.assertAlmostEqual(uc.convert_temp_F_to_R(50.0), 509.67, places=5)

        with self.assertRaises(UnitsError):
            uc.convert_temp_K_to_C(ex)

    def test_module_example(self):
        from pyomo.environ import ConcreteModel, Var, Objective, units

        model = ConcreteModel()
        model.acc = Var()
        model.obj = Objective(
            expr=(model.acc * units.m / units.s**2 - 9.81 * units.m / units.s**2) ** 2
        )
        self.assertEqual('m**2/s**4', str(units.get_units(model.obj.expr)))

    def test_convert_value(self):
        u = units
        x = 0.4535923
        expected_lb_value = 1.0
        actual_lb_value = u.convert_value(num_value=x, from_units=u.kg, to_units=u.lb)
        self.assertAlmostEqual(expected_lb_value, actual_lb_value, places=5)
        actual_lb_value = u.convert_value(
            num_value=value(x * u.kg), from_units=u.kg, to_units=u.lb
        )
        self.assertAlmostEqual(expected_lb_value, actual_lb_value, places=5)

        src = 5
        ans = u.convert_value(src, u.m / u.s, u.m / u.s)
        self.assertIs(src, ans)

        with self.assertRaises(UnitsError):
            # cannot convert from meters to pounds
            actual_lb_value = u.convert_value(
                num_value=x, from_units=u.meters, to_units=u.lb
            )

        with self.assertRaises(UnitsError):
            # num_value must be a native numerical type
            actual_lb_value = u.convert_value(
                num_value=x * u.kg, from_units=u.kg, to_units=u.lb
            )

    def test_convert(self):
        u = units
        m = ConcreteModel()
        m.dx = Var(units=u.m, initialize=0.10188943773836046)
        m.dy = Var(units=u.m, initialize=0.0)
        m.vx = Var(units=u.m / u.s, initialize=0.7071067769802851)
        m.vy = Var(units=u.m / u.s, initialize=0.7071067769802851)
        m.t = Var(units=u.min, bounds=(1e-5, 10.0), initialize=0.0024015570927624456)
        m.theta = Var(
            bounds=(0, 0.49 * 3.14), initialize=0.7853981693583533, units=u.radians
        )
        m.a = Param(initialize=-32.2, units=u.ft / u.s**2)

        m.obj = Objective(expr=m.dx, sense=maximize)
        m.vx_con = Constraint(expr=m.vx == 1.0 * u.m / u.s * cos(m.theta))
        m.vy_con = Constraint(expr=m.vy == 1.0 * u.m / u.s * sin(m.theta))
        m.dx_con = Constraint(expr=m.dx == m.vx * u.convert(m.t, to_units=u.s))
        m.dy_con = Constraint(
            expr=m.dy
            == m.vy * u.convert(m.t, to_units=u.s)
            + 0.5
            * (u.convert(m.a, to_units=u.m / u.s**2))
            * (u.convert(m.t, to_units=u.s)) ** 2
        )
        m.ground = Constraint(expr=m.dy == 0)

        with self.assertRaises(UnitsError):
            u.convert(m.a, to_units=u.kg)

        self.assertAlmostEqual(value(m.obj), 0.10188943773836046, places=5)
        self.assertAlmostEqual(value(m.vx_con.body), 0.0, places=5)
        self.assertAlmostEqual(value(m.vy_con.body), 0.0, places=5)
        self.assertAlmostEqual(value(m.dx_con.body), 0.0, places=5)
        self.assertAlmostEqual(value(m.dy_con.body), 0.0, places=5)
        self.assertAlmostEqual(value(m.ground.body), 0.0, places=5)

    def test_convert_dimensionless(self):
        u = units
        m = ConcreteModel()
        m.x = Var()
        foo = u.convert(m.x, to_units=u.dimensionless)
        foo = u.convert(m.x, to_units=None)
        foo = u.convert(m.x, to_units=1.0)
        with self.assertRaises(InconsistentUnitsError):
            foo = u.convert(m.x, to_units=u.kg)
        m.y = Var(units=u.kg)
        with self.assertRaises(InconsistentUnitsError):
            foo = u.convert(m.y, to_units=u.dimensionless)
        with self.assertRaises(InconsistentUnitsError):
            foo = u.convert(m.y, to_units=None)
        with self.assertRaises(InconsistentUnitsError):
            foo = u.convert(m.y, to_units=1.0)

    def test_usd(self):
        u = units
        u.load_definitions_from_strings(["USD = [currency]"])
        expr = 3.0 * u.USD
        self._get_check_units_ok(expr, u, 'USD')

    def test_clone(self):
        m = ConcreteModel()
        m.x = Var(units=units.kg)
        m.c = Constraint(expr=m.x**2 <= 10 * units.kg**2)
        i = m.clone()
        self.assertIs(m.x._units, i.x._units)
        self.assertEqual(str(m.c.upper), str(i.c.upper))
        base = StringIO()
        m.pprint(base)
        test = StringIO()
        i.pprint(test)
        self.assertEqual(base.getvalue(), test.getvalue())

    def test_pickle(self):
        m = ConcreteModel()
        m.x = Var(units=units.kg)
        m.c = Constraint(expr=m.x**2 <= 10 * units.kg**2)
        log = StringIO()
        with LoggingIntercept(log, 'pyomo.core.base'):
            i = pickle.loads(pickle.dumps(m))
        self.assertEqual("", log.getvalue())
        self.assertIsNot(m.x, i.x)
        self.assertIsNot(m.x._units, i.x._units)
        self.assertTrue(check_units_equivalent(m.x._units, i.x._units))
        self.assertEqual(str(m.c.upper), str(i.c.upper))
        base = StringIO()
        m.pprint(base)
        test = StringIO()
        i.pprint(test)
        self.assertEqual(base.getvalue(), test.getvalue())

        # Test pickling a custom units manager
        um = PyomoUnitsContainer(pint_module.UnitRegistry())
        m = ConcreteModel()
        m.x = Var(units=um.kg)
        m.c = Constraint(expr=m.x**2 <= 10 * um.kg**2)
        log = StringIO()
        with LoggingIntercept(log, 'pyomo.core.base'):
            i = pickle.loads(pickle.dumps(m))
        self.assertIn(
            "pickling a _PyomoUnit associated with a PyomoUnitsContainer "
            "that is not the default singleton "
            "(pyomo.core.base.units_container.units)",
            log.getvalue(),
        )
        self.assertIsNot(m.x, i.x)
        self.assertIsNot(m.x._units, i.x._units)
        # Note that pint is inconsistent when comparing standard units
        # across different UnitRegistry instances: older versions of
        # pint would have them compare "not equal" while newer versions
        # compare equal
        #
        # self.assertNotEqual(m.x._units, i.x._units)
        self.assertEqual(str(m.c.upper), str(i.c.upper))
        base = StringIO()
        m.pprint(base)
        test = StringIO()
        i.pprint(test)
        self.assertEqual(base.getvalue(), test.getvalue())

    def test_set_pint_registry(self):
        um = _DeferredUnitsSingleton()
        pint_reg = pint_module.UnitRegistry()
        # Test we can (silently) set the registry if it is the first
        # thing we do
        with LoggingIntercept() as LOG:
            um.set_pint_registry(pint_reg)
        self.assertEqual(LOG.getvalue(), "")
        self.assertIs(um.pint_registry, pint_reg)
        # Test that a no-op set is silent
        with LoggingIntercept() as LOG:
            um.set_pint_registry(pint_reg)
        self.assertEqual(LOG.getvalue(), "")
        # Test that changing the registry generates a warning
        with LoggingIntercept() as LOG:
            um.set_pint_registry(pint_module.UnitRegistry())
        self.assertIn(
            "Changing the pint registry used by the Pyomo Units "
            "system after the PyomoUnitsContainer was constructed",
            LOG.getvalue(),
        )

    def test_as_quantity_scalar(self):
        _pint = units._pint_registry
        Quantity = _pint.Quantity
        m = ConcreteModel()
        m.x = Var(initialize=1)
        m.y = Var(initialize=2, units=units.g)
        m.p = Param(initialize=3)
        m.q = Param(initialize=4, units=1 / units.s)
        m.b = BooleanVar(initialize=True)

        q = as_quantity(0)
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 0 * _pint.dimensionless)

        q = as_quantity(None)
        self.assertIs(q.__class__, None.__class__)
        self.assertEqual(q, None)

        q = as_quantity(str('aaa'))
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 'aaa' * _pint.dimensionless)

        q = as_quantity(True)
        self.assertIs(q.__class__, bool)
        self.assertEqual(q, True)

        q = as_quantity(units.kg)
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 1 * _pint.kg)

        q = as_quantity(NumericConstant(5))
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 5 * _pint.dimensionless)

        q = as_quantity(m.x)
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 1 * _pint.dimensionless)

        q = as_quantity(m.y)
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 2 * _pint.g)

        q = as_quantity(m.p)
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 3 * _pint.dimensionless)

        q = as_quantity(m.q)
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 4 / _pint.s)

        q = as_quantity(m.b)
        self.assertIs(q.__class__, bool)
        self.assertEqual(q, True)

        class UnknownPyomoType:
            def is_expression_type(self, expression_system=None):
                return False

            def is_numeric_type(self):
                return False

            def is_logical_type(self):
                return False

        other = UnknownPyomoType()
        q = as_quantity(other)
        self.assertIs(q.__class__, UnknownPyomoType)
        self.assertIs(q, other)

    def test_as_quantity_expression(self):
        _pint = units._pint_registry
        Quantity = _pint.Quantity
        m = ConcreteModel()
        m.x = Var(initialize=1)
        m.y = Var(initialize=2, units=units.g)
        m.p = Param(initialize=3)
        m.q = Param(initialize=4, units=1 / units.s)

        q = as_quantity(m.x * m.p)
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 3 * _pint.dimensionless)

        q = as_quantity(m.x * m.q)
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 4 / _pint.s)

        q = as_quantity(m.y * m.p)
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 6 * _pint.g)

        q = as_quantity(m.y * m.q)
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 8 * _pint.g / _pint.s)

        q = as_quantity(m.y <= 2 * m.y)
        self.assertIs(q.__class__, bool)
        self.assertEqual(q, True)

        q = as_quantity(m.y >= 2 * m.y)
        self.assertIs(q.__class__, bool)
        self.assertEqual(q, False)

        q = as_quantity(EXPR.Expr_if(IF=m.y <= 2 * m.y, THEN=m.x, ELSE=m.p))
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 1 * _pint.dimensionless)

        q = as_quantity(EXPR.Expr_if(IF=m.y >= 2 * m.y, THEN=m.x, ELSE=m.p))
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 3 * _pint.dimensionless)

        # NOTE: The following two tests are not unit consistent (but can
        # be evaluated)

        q = as_quantity(EXPR.Expr_if(IF=m.x <= 2 * m.x, THEN=m.y, ELSE=m.q))
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 2 * _pint.g)

        q = as_quantity(EXPR.Expr_if(IF=m.x >= 2 * m.x, THEN=m.y, ELSE=m.q))
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q, 4 / _pint.s)

        # Note: check the units explicitly, as
        #   Quantity(x, radian) == Quantity(x, dimensionless)
        q = as_quantity(acos(m.x))
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q.units, _pint.radian)
        self.assertEqual(q, 0 * _pint.radian)

        q = as_quantity(cos(m.x * math.pi))
        self.assertIs(q.__class__, Quantity)
        self.assertEqual(q.units, _pint.dimensionless)
        self.assertAlmostEqual(q, -1 * _pint.dimensionless)

        def MyAdder(x, y):
            return x + y

        m.EF = ExternalFunction(MyAdder, units=units.kg)
        ef = m.EF(m.x, m.y)
        q = as_quantity(ef)
        self.assertIs(q.__class__, Quantity)
        self.assertAlmostEqual(q, 3 * _pint.kg)

    def test_var_set_value(self):
        # Tests for #1570
        m = ConcreteModel()

        # Un-united variables just strip the units
        m.x = Var()
        m.x.value = 10
        self.assertEqual(m.x.value, 10)
        m.x.value = 20 * units.kg
        self.assertEqual(m.x.value, 20)
        m.x.value = 30 * units.dimensionless
        self.assertEqual(m.x.value, 30)
        del m.x

        # Dimensionless variables require dimensionless args
        m.x = Var(units=units.dimensionless)
        m.x.value = 10
        self.assertEqual(m.x.value, 10)
        with self.assertRaisesRegex(UnitsError, 'Cannot convert kg to dimensionless'):
            m.x.value = 20 * units.kg
        self.assertEqual(m.x.value, 10)
        m.x.value = 30 * units.dimensionless
        self.assertEqual(m.x.value, 30)
        del m.x

        # Dimensioned variables require bare or dimensioned args
        m.x = Var(units=units.gram)
        m.x.value = 10
        self.assertEqual(m.x.value, 10)
        m.x.value = 20 * units.kg
        self.assertEqual(m.x.value, 20000)
        with self.assertRaisesRegex(UnitsError, 'Cannot convert dimensionless to g'):
            m.x.value = 30 * units.dimensionless
        self.assertEqual(m.x.value, 20000)
        del m.x


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for indexed components
#

import os
from os.path import abspath, dirname

currdir = dirname(abspath(__file__)) + os.sep

from pyomo.common import DeveloperError
import pyomo.common.unittest as unittest
from pyomo.common.log import LoggingIntercept

from pyomo.environ import ConcreteModel, Var, Param, Set, value, Integers
from pyomo.core.base.set import FiniteSetOf, OrderedSetOf
from pyomo.core.base.indexed_component import normalize_index
from pyomo.core.expr import GetItemExpression
from pyomo.core import SortComponents


class TestSimpleVar(unittest.TestCase):
    def test0(self):
        # Test fixed attribute - 1D
        m = ConcreteModel()
        m.x = Var()

        names = set()
        for var in m.x[:]:
            names.add(var.name)
        self.assertEqual(names, set(['x']))

    def test1(self):
        # Test fixed attribute - 1D
        m = ConcreteModel()
        m.x = Var(range(3), dense=True)

        names = set()
        for var in m.x[:]:
            names.add(var.name)
        self.assertEqual(names, set(['x[0]', 'x[1]', 'x[2]']))

    def test2a(self):
        # Test fixed attribute - 2D
        m = ConcreteModel()
        m.x = Var(range(3), range(3), dense=True)

        names = set()
        for var in m.x[:, 1]:
            names.add(var.name)
        self.assertEqual(names, set(['x[0,1]', 'x[1,1]', 'x[2,1]']))

    def test2b(self):
        # Test fixed attribute - 2D
        m = ConcreteModel()
        m.x = Var(range(3), range(3), dense=True)

        names = set()
        for var in m.x[2, :]:
            names.add(var.name)
        self.assertEqual(names, set(['x[2,0]', 'x[2,1]', 'x[2,2]']))

    def test2c(self):
        # Test fixed attribute - 2D
        m = ConcreteModel()
        m.x = Var(range(3), range(3), dense=True)

        names = set()
        for var in m.x[3, :]:
            names.add(var.name)
        self.assertEqual(names, set())

    def test3a(self):
        # Test fixed attribute - 3D
        m = ConcreteModel()
        m.x = Var(range(3), range(3), range(3), dense=True)

        names = set()
        for var in m.x[:, 1, :]:
            names.add(var.name)
        self.assertEqual(
            names,
            set(
                [
                    'x[0,1,0]',
                    'x[0,1,1]',
                    'x[0,1,2]',
                    'x[1,1,0]',
                    'x[1,1,1]',
                    'x[1,1,2]',
                    'x[2,1,0]',
                    'x[2,1,1]',
                    'x[2,1,2]',
                ]
            ),
        )

    def test3b(self):
        # Test fixed attribute - 3D
        m = ConcreteModel()
        m.x = Var(range(3), range(3), range(3), dense=True)

        names = set()
        for var in m.x[0, :, 2]:
            names.add(var.name)
        self.assertEqual(names, set(['x[0,0,2]', 'x[0,1,2]', 'x[0,2,2]']))


class TestIndexedComponent(unittest.TestCase):
    def test_normalize_index(self):
        # Test that normalize_index works correctly
        self.assertEqual("abc", normalize_index("abc"))
        self.assertEqual(1, normalize_index(1))
        self.assertEqual(1, normalize_index([1]))
        self.assertEqual((1, 2, 3), normalize_index((1, 2, 3)))
        self.assertEqual((1, 2, 3), normalize_index([1, 2, 3]))
        self.assertEqual((1, 2, 3, 4), normalize_index((1, 2, [3, 4])))
        self.assertEqual((1, 2, 'abc'), normalize_index((1, 2, 'abc')))
        self.assertEqual((1, 2, 'abc'), normalize_index((1, 2, ('abc',))))
        a = [0, 9, 8]
        self.assertEqual((1, 2, 0, 9, 8), normalize_index((1, 2, a)))
        self.assertEqual(
            (1, 2, 3, 4, 5),
            normalize_index([[], 1, [], 2, [[], 3, [[], 4, []], []], 5, []]),
        )
        self.assertEqual((), normalize_index([[[[], []], []], []]))
        self.assertEqual((), normalize_index([[], [[], [[]]]]))

        # Test that normalize_index doesn't expand component-like things
        m = ConcreteModel()
        m.x = Var()
        m.y = Var([1])
        m.i = Set(initialize=[1])
        m.j = Set([1], initialize=[1])
        self.assertIs(m, normalize_index(m))
        self.assertIs(m.x, normalize_index(m.x))
        self.assertIs(m.y, normalize_index(m.y))
        self.assertIs(m.y[1], normalize_index(m.y[1]))
        self.assertIs(m.i, normalize_index(m.i))
        self.assertIs(m.j, normalize_index(m.j))
        self.assertIs(m.j[1], normalize_index(m.j[1]))

    def test_index_by_constant_simpleComponent(self):
        m = ConcreteModel()
        m.i = Param(initialize=2)
        m.x = Var([1, 2, 3], initialize=lambda m, x: 2 * x)
        self.assertEqual(value(m.x[2]), 4)
        self.assertEqual(value(m.x[m.i]), 4)
        self.assertIs(m.x[2], m.x[m.i])

    def test_index_by_multiple_constant_simpleComponent(self):
        m = ConcreteModel()
        m.i = Param(initialize=2)
        m.j = Param(initialize=3)
        m.x = Var([1, 2, 3], [1, 2, 3], initialize=lambda m, x, y: 2 * x * y)
        self.assertEqual(value(m.x[2, 3]), 12)
        self.assertEqual(value(m.x[m.i, 3]), 12)
        self.assertEqual(value(m.x[m.i, m.j]), 12)
        self.assertEqual(value(m.x[2, m.j]), 12)
        self.assertIs(m.x[2, 3], m.x[m.i, 3])
        self.assertIs(m.x[2, 3], m.x[m.i, m.j])
        self.assertIs(m.x[2, 3], m.x[2, m.j])

    def test_index_by_fixed_simpleComponent(self):
        m = ConcreteModel()
        m.i = Param(initialize=2, mutable=True)
        m.x = Var([1, 2, 3], initialize=lambda m, x: 2 * x)
        self.assertEqual(value(m.x[2]), 4)
        self.assertRaisesRegex(
            RuntimeError, 'is a fixed but not constant value', m.x.__getitem__, m.i
        )

    def test_index_by_variable_simpleComponent(self):
        m = ConcreteModel()
        m.i = Var(initialize=2, domain=Integers)
        m.x = Var([1, 2, 3], initialize=lambda m, x: 2 * x)
        self.assertEqual(value(m.x[2]), 4)

        # Test we can index by a variable
        thing = m.x[m.i]
        self.assertIsInstance(thing, GetItemExpression)
        self.assertEqual(len(thing.args), 2)
        self.assertIs(thing.args[0], m.x)
        self.assertIs(thing.args[1], m.i)

        # Test we can index by an integer-valued expression
        idx_expr = 2 * m.i + 1
        thing = m.x[idx_expr]
        self.assertIsInstance(thing, GetItemExpression)
        self.assertEqual(len(thing.args), 2)
        self.assertIs(thing.args[0], m.x)
        self.assertIs(thing.args[1], idx_expr)

    def test_index_param_by_variable(self):
        m = ConcreteModel()
        m.i = Var(initialize=2, domain=Integers)
        m.p = Param([1, 2, 3], initialize=lambda m, x: 2 * x)

        # Test we can index by a variable
        thing = m.p[m.i]
        self.assertIsInstance(thing, GetItemExpression)
        self.assertEqual(len(thing.args), 2)
        self.assertIs(thing.args[0], m.p)
        self.assertIs(thing.args[1], m.i)

        # Test we can index by an integer-valued expression
        idx_expr = 2**m.i + 1
        thing = m.p[idx_expr]
        self.assertIsInstance(thing, GetItemExpression)
        self.assertEqual(len(thing.args), 2)
        self.assertIs(thing.args[0], m.p)
        self.assertIs(thing.args[1], idx_expr)

    def test_index_var_by_tuple_with_variables(self):
        m = ConcreteModel()
        m.x = Var([(1, 1), (2, 1), (1, 2), (2, 2)])
        m.i = Var([1, 2, 3], domain=Integers)

        thing = m.x[1, m.i[1]]
        self.assertIsInstance(thing, GetItemExpression)
        self.assertEqual(len(thing.args), 3)
        self.assertIs(thing.args[0], m.x)
        self.assertEqual(thing.args[1], 1)
        self.assertIs(thing.args[2], m.i[1])

        idx_expr = m.i[1] + m.i[2] * m.i[3]
        thing = m.x[1, idx_expr]
        self.assertIsInstance(thing, GetItemExpression)
        self.assertEqual(len(thing.args), 3)
        self.assertIs(thing.args[0], m.x)
        self.assertEqual(thing.args[1], 1)
        self.assertIs(thing.args[2], idx_expr)

    def test_index_by_unhashable_type(self):
        m = ConcreteModel()
        m.x = Var([1, 2, 3], initialize=lambda m, x: 2 * x)
        # Indexing by a dict raises an error
        self.assertRaisesRegex(TypeError, '.*', m.x.__getitem__, {})
        # Indexing by lists works...
        # ... scalar
        self.assertIs(m.x[[1]], m.x[1])
        # ... "tuple"
        m.y = Var([(1, 1), (1, 2)])
        self.assertIs(m.y[[1, 1]], m.y[1, 1])
        m.y[[1, 2]] = 5
        y12 = m.y[[1, 2]]
        self.assertEqual(y12.value, 5)
        m.y[[1, 2]] = 15
        self.assertIs(y12, m.y[[1, 2]])
        self.assertEqual(y12.value, 15)
        with self.assertRaisesRegex(
            KeyError, r"Index '\(2, 2\)' is not valid for indexed component 'y'"
        ):
            m.y[[2, 2]] = 5

    def test_ordered_keys(self):
        m = ConcreteModel()
        # Pick a set whose unordered iteration order should never match
        # the "ordered" iteration order.
        init_keys = [2, 1, (1, 2), (1, 'a'), (1, 1)]
        m.I = Set(ordered=False, dimen=None, initialize=init_keys)
        ordered_keys = [1, 2, (1, 1), (1, 2), (1, 'a')]
        m.x = Var(m.I)
        self.assertNotEqual(list(m.x.keys()), list(m.x.keys(True)))
        self.assertEqual(set(m.x.keys()), set(m.x.keys(True)))
        self.assertEqual(ordered_keys, list(m.x.keys(True)))

        m.P = Param(m.I, initialize={k: v for v, k in enumerate(init_keys)})
        self.assertNotEqual(list(m.P.keys()), list(m.P.keys(True)))
        self.assertEqual(set(m.P.keys()), set(m.P.keys(True)))
        self.assertEqual(ordered_keys, list(m.P.keys(True)))
        self.assertEqual([1, 0, 4, 2, 3], list(m.P.values(True)))
        self.assertEqual(
            list(zip(ordered_keys, [1, 0, 4, 2, 3])), list(m.P.items(True))
        )

        m.P = Param(m.I, initialize={(1, 2): 30, 1: 10, 2: 20}, default=1)
        self.assertNotEqual(list(m.P.keys()), list(m.P.keys(True)))
        self.assertEqual(set(m.P.keys()), set(m.P.keys(True)))
        self.assertEqual(ordered_keys, list(m.P.keys(True)))
        self.assertEqual([10, 20, 1, 30, 1], list(m.P.values(True)))
        self.assertEqual(
            list(zip(ordered_keys, [10, 20, 1, 30, 1])), list(m.P.items(True))
        )

    def test_ordered_keys_deprecation(self):
        m = ConcreteModel()
        unordered = [1, 3, 2]
        ordered = [1, 2, 3]
        m.I = FiniteSetOf(unordered)
        m.x = Var(m.I)
        self.assertEqual(list(m.x.keys()), unordered)
        self.assertEqual(list(m.x.keys(SortComponents.ORDERED_INDICES)), ordered)
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.keys(True)), ordered)
        self.assertEqual(LOG.getvalue(), "")
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.keys(ordered=True)), ordered)
        self.assertIn('keys(ordered=True) is deprecated', LOG.getvalue())
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.keys(ordered=False)), unordered)
        self.assertIn('keys(ordered=False) is deprecated', LOG.getvalue())

        m = ConcreteModel()
        unordered = [1, 3, 2]
        ordered = [1, 2, 3]
        m.I = OrderedSetOf(unordered)
        m.x = Var(m.I)
        self.assertEqual(list(m.x.keys()), unordered)
        self.assertEqual(list(m.x.keys(SortComponents.ORDERED_INDICES)), unordered)
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.keys(True)), ordered)
        self.assertEqual(LOG.getvalue(), "")
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.keys(ordered=True)), unordered)
        self.assertIn('keys(ordered=True) is deprecated', LOG.getvalue())
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.keys(ordered=False)), unordered)
        self.assertIn('keys(ordered=False) is deprecated', LOG.getvalue())

    def test_ordered_values_deprecation(self):
        m = ConcreteModel()
        unordered = [1, 3, 2]
        ordered = [1, 2, 3]
        m.I = FiniteSetOf(unordered)
        m.x = Var(m.I)
        unordered = [m.x[i] for i in unordered]
        ordered = [m.x[i] for i in ordered]
        self.assertEqual(list(m.x.values()), unordered)
        self.assertEqual(list(m.x.values(SortComponents.ORDERED_INDICES)), ordered)
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.values(True)), ordered)
        self.assertEqual(LOG.getvalue(), "")
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.values(ordered=True)), ordered)
        self.assertIn('values(ordered=True) is deprecated', LOG.getvalue())
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.values(ordered=False)), unordered)
        self.assertIn('values(ordered=False) is deprecated', LOG.getvalue())

        m = ConcreteModel()
        unordered = [1, 3, 2]
        ordered = [1, 2, 3]
        m.I = OrderedSetOf(unordered)
        m.x = Var(m.I)
        unordered = [m.x[i] for i in unordered]
        ordered = [m.x[i] for i in ordered]
        self.assertEqual(list(m.x.values()), unordered)
        self.assertEqual(list(m.x.values(SortComponents.ORDERED_INDICES)), unordered)
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.values(True)), ordered)
        self.assertEqual(LOG.getvalue(), "")
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.values(ordered=True)), unordered)
        self.assertIn('values(ordered=True) is deprecated', LOG.getvalue())
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.values(ordered=False)), unordered)
        self.assertIn('values(ordered=False) is deprecated', LOG.getvalue())

    def test_ordered_items_deprecation(self):
        m = ConcreteModel()
        unordered = [1, 3, 2]
        ordered = [1, 2, 3]
        m.I = FiniteSetOf(unordered)
        m.x = Var(m.I)
        unordered = [(i, m.x[i]) for i in unordered]
        ordered = [(i, m.x[i]) for i in ordered]
        self.assertEqual(list(m.x.items()), unordered)
        self.assertEqual(list(m.x.items(SortComponents.ORDERED_INDICES)), ordered)
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.items(True)), ordered)
        self.assertEqual(LOG.getvalue(), "")
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.items(ordered=True)), ordered)
        self.assertIn('items(ordered=True) is deprecated', LOG.getvalue())
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.items(ordered=False)), unordered)
        self.assertIn('items(ordered=False) is deprecated', LOG.getvalue())

        m = ConcreteModel()
        unordered = [1, 3, 2]
        ordered = [1, 2, 3]
        m.I = OrderedSetOf(unordered)
        m.x = Var(m.I)
        unordered = [(i, m.x[i]) for i in unordered]
        ordered = [(i, m.x[i]) for i in ordered]
        self.assertEqual(list(m.x.items()), unordered)
        self.assertEqual(list(m.x.items(SortComponents.ORDERED_INDICES)), unordered)
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.items(True)), ordered)
        self.assertEqual(LOG.getvalue(), "")
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.items(ordered=True)), unordered)
        self.assertIn('items(ordered=True) is deprecated', LOG.getvalue())
        with LoggingIntercept() as LOG:
            self.assertEqual(list(m.x.items(ordered=False)), unordered)
        self.assertIn('items(ordered=False) is deprecated', LOG.getvalue())

    def test_index_attribute_out_of_sync(self):
        m = ConcreteModel()
        m.x = Var([1, 2, 3])
        # make sure everything is right to begin with
        for i in [1, 2, 3]:
            self.assertEqual(m.x[i].index(), i)
        # now mess it up
        m.x[3]._index = 2
        with self.assertRaisesRegex(
            DeveloperError,
            ".*The '_data' dictionary and '_index' attribute are out of "
            "sync for indexed Var 'x': The 2 entry in the '_data' "
            "dictionary does not map back to this component data object.",
            normalize_whitespace=True,
        ):
            m.x[3].index()


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import pyomo.common.unittest as unittest

from pyomo.core.base.indexed_component import normalize_index
from pyomo.core.base.indexed_component_slice import IndexedComponent_slice
from pyomo.core.base.global_set import UnindexedComponent_set
import pyomo.environ as pyo
import pyomo.dae as dae
from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.util.slices import (
    get_component_call_stack,
    slice_component_along_sets,
    replace_indices,
    get_location_set_map,
)


class TestGetComponentCallStack(unittest.TestCase):
    get_attribute = IndexedComponent_slice.get_attribute
    get_item = IndexedComponent_slice.get_item

    def assertSameStack(self, stack1, stack2):
        for (call1, arg1), (call2, arg2) in zip(stack1, stack2):
            self.assertIs(call1, call2)
            self.assertEqual(arg1, arg2)

    def assertCorrectStack(self, comp, pred_stack, context=None):
        act_stack = get_component_call_stack(comp, context=None)
        self.assertSameStack(pred_stack, act_stack)

    def model(self):
        m = pyo.ConcreteModel()
        m.s1 = pyo.Set(initialize=[1, 2, 3])
        m.s2 = pyo.Set(initialize=[('a', 1), ('b', 2)])

        m.v0 = pyo.Var()
        m.v1 = pyo.Var(m.s1)
        m.v2 = pyo.Var(m.s2)

        @m.Block()
        def b(b):
            @b.Block(m.s1)
            def b1(b):
                b.v0 = pyo.Var()
                b.v1 = pyo.Var(m.s1)

                @b.Block()
                def b(b):
                    b.v0 = pyo.Var()
                    b.v2 = pyo.Var(m.s2)

                    @b.Block(m.s1, m.s2)
                    def b2(b, i1, i2, i3=None):
                        # Optional i3 is to allow for the case
                        # when normalize_index.flatten is False
                        b.v0 = pyo.Var()
                        b.v1 = pyo.Var(m.s1)

                        @b.Block()
                        def b(b):
                            b.v0 = pyo.Var()
                            b.v2 = pyo.Var(m.s2)

        return m

    def test_no_context(self):
        m = self.model()

        comp = m.v1[1]
        pred_stack = [(self.get_item, 1), (self.get_attribute, 'v1')]
        self.assertCorrectStack(comp, pred_stack)

        comp = m.v1
        pred_stack = [(self.get_attribute, 'v1')]
        self.assertCorrectStack(comp, pred_stack)

        comp = m.b.b1[1].b.b2
        pred_stack = [
            (self.get_attribute, 'b2'),
            (self.get_attribute, 'b'),
            (self.get_item, 1),
            (self.get_attribute, 'b1'),
            (self.get_attribute, 'b'),
        ]
        self.assertCorrectStack(comp, pred_stack)

        comp = m.b.b1[1].b.b2[1, 'a', 1]
        pred_stack = [
            (self.get_item, (1, 'a', 1)),
            (self.get_attribute, 'b2'),
            (self.get_attribute, 'b'),
            (self.get_item, 1),
            (self.get_attribute, 'b1'),
            (self.get_attribute, 'b'),
        ]
        self.assertCorrectStack(comp, pred_stack)

        normalize_index.flatten = False
        comp = m.b.b1[1].b.b2[1, ('a', 1)]
        # NOTE: This fails without the parenthesis around ('a',1)
        # as comp.index() is None. The actual stack just starts one
        # layer higher...
        # It is confusing that comp can even be generated in this case;
        # shouldn't (1,'a',1) be an invalid index?
        pred_stack = [
            (self.get_item, (1, ('a', 1))),
            (self.get_attribute, 'b2'),
            (self.get_attribute, 'b'),
            (self.get_item, 1),
            (self.get_attribute, 'b1'),
            (self.get_attribute, 'b'),
        ]
        self.assertCorrectStack(comp, pred_stack)
        normalize_index.flatten = True

    def test_from_block(self):
        m = self.model()

        comp = m.v0
        pred_stack = [(self.get_attribute, 'v0')]
        self.assertCorrectStack(comp, pred_stack, context=m)

        comp = m.b.b1[2].b.b2[1, 'a', 1]
        pred_stack = [
            (self.get_item, (1, 'a', 1)),
            (self.get_attribute, 'b2'),
            (self.get_attribute, 'b'),
            (self.get_item, 2),
            (self.get_attribute, 'b1'),
            (self.get_attribute, 'b'),
        ]
        self.assertCorrectStack(comp, pred_stack, context=m)

        comp = m.b.b1[2].b.b2[1, 'a', 1].b.v2['b', 2]
        pred_stack = [
            (self.get_item, ('b', 2)),
            (self.get_attribute, 'v2'),
            (self.get_attribute, 'b'),
            (self.get_item, (1, 'a', 1)),
            (self.get_attribute, 'b2'),
            (self.get_attribute, 'b'),
            (self.get_item, 2),
        ]
        self.assertCorrectStack(comp, pred_stack, context=m.b.b1)

        comp = m.b.b1[2].b.b2
        pred_stack = [
            (self.get_attribute, 'b2'),
            (self.get_attribute, 'b'),
            (self.get_item, 2),
        ]
        self.assertCorrectStack(comp, pred_stack, context=m.b.b1)

        comp = m.b.b1[2]
        pred_stack = [(self.get_item, 2)]
        self.assertCorrectStack(comp, pred_stack, context=m.b.b1)

        comp = m.b.b1
        act_stack = get_component_call_stack(comp, context=m.b.b1)
        self.assertEqual(len(act_stack), 0)

    def test_from_blockdata(self):
        m = self.model()

        context = m.b.b1[3].b.b2[2, 'b', 2]
        comp = m.b.b1[3].b.b2[2, 'b', 2].b
        pred_stack = [(self.get_attribute, 'b')]
        self.assertCorrectStack(comp, pred_stack, context=context)

        comp = m.b.b1[3].b.b2[2, 'b', 2].b.v2['a', 1]
        pred_stack = [
            (self.get_item, ('a', 1)),
            (self.get_attribute, 'v2'),
            (self.get_attribute, 'b'),
        ]
        self.assertCorrectStack(comp, pred_stack, context=context)

        context = m.b.b1[3]
        comp = m.b.b1[3]
        act_stack = get_component_call_stack(comp, context=context)
        self.assertEqual(len(act_stack), 0)


class TestGetLocationAndReplacement(unittest.TestCase):
    def model(self):
        m = pyo.ConcreteModel()

        m.time = pyo.Set(initialize=[1, 2, 3])
        m.space = dae.ContinuousSet(initialize=[0, 2])
        m.comp = pyo.Set(initialize=['a', 'b'])
        m.d_2 = pyo.Set(initialize=[('a', 1), ('b', 2)])
        m.d_none = pyo.Set(initialize=[('c', 1, 10), ('d', 3)], dimen=None)

        m.b0 = pyo.Block()
        m.b1 = pyo.Block(m.time)
        m.b2 = pyo.Block(m.time, m.space)
        m.b0d2 = pyo.Block(m.d_2)
        m.b1d2 = pyo.Block(m.time, m.d_2)
        m.b2d2 = pyo.Block(m.time, m.d_2, m.space)
        m.b3d2 = pyo.Block(m.time, m.d_2, m.space, m.time)

        m.b0dn = pyo.Block(m.d_none)
        m.b1dn = pyo.Block(m.time, m.d_none)
        m.b1dnd2 = pyo.Block(m.time, m.d_none, m.d_2)
        m.dnd2b1 = pyo.Block(m.d_none, m.d_2, m.time)
        m.b2dn = pyo.Block(m.time, m.space, m.d_none)
        m.b2dnd2 = pyo.Block(m.time, m.d_none, m.d_2, m.space)
        m.b3dn = pyo.Block(m.time, m.d_2, m.d_none, m.space, m.d_2)

        m.dn2 = pyo.Block(m.time, m.d_none, m.d_2, m.d_none, m.d_2)

        return m

    def assertSameMap(self, pred, act):
        self.assertEqual(len(pred), len(act))
        for k, v in pred.items():
            self.assertIn(k, act)
            self.assertIs(pred[k], act[k])

    def test_one_to_one(self):
        m = self.model()

        index_set = m.b0.index_set()
        index = None
        pred_map = {0: UnindexedComponent_set}
        location_set_map = get_location_set_map(index, index_set)
        self.assertEqual(pred_map, location_set_map)

        index_set = m.b1.index_set()
        index = 1
        pred_map = {0: m.time}
        location_set_map = get_location_set_map(index, index_set)
        self.assertSameMap(pred_map, location_set_map)

        index_set = m.b2.index_set()
        index = (1, 2)
        pred_map = {0: m.time, 1: m.space}
        location_set_map = get_location_set_map(index, index_set)
        self.assertSameMap(pred_map, location_set_map)

    def test_1d_replacement(self):
        m = self.model()

        # Unindexed
        index_set = m.b0.index_set()
        index = None
        location_set_map = get_location_set_map(index, index_set)
        sets = ComponentSet()
        pred_index = (None,)
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(pred_index, new_index)

        # One one-d set
        index_set = m.b1.index_set()
        index = 1
        location_set_map = get_location_set_map(index, index_set)
        sets = ComponentSet()
        pred_index = (1,)
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(pred_index, new_index)

        sets = ComponentSet((m.time,))
        pred_index = (slice(None),)
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(pred_index, new_index)

        # Two one-d sets
        index_set = m.b2.index_set()
        index = (1, 2)
        location_set_map = get_location_set_map(index, index_set)
        sets = ComponentSet()
        pred_index = (1, 2)
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(pred_index, new_index)

        sets = ComponentSet((m.space,))
        pred_index = (1, slice(None))
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(pred_index, new_index)

        sets = ComponentSet((m.space, m.time))
        pred_index = (slice(None), slice(None))
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(pred_index, new_index)

    def test_multi_dim(self):
        m = self.model()
        index_set = m.b0d2.index_set()
        index = ('a', 1)
        pred_map = {0: m.d_2, 1: m.d_2}
        location_set_map = get_location_set_map(index, index_set)
        self.assertSameMap(pred_map, location_set_map)

        normalize_index.flatten = False
        # A single multi-dimensional index will behave the same when
        # normalize_index.flatten is False. This is probably unexpected
        # Should we just fail in this case?
        # Note that the index (('a',1),) will not work.
        index_set = m.b0d2.index_set()
        index = ('a', 1)
        pred_map = {0: m.d_2, 1: m.d_2}
        with self.assertRaises(ValueError) as cm:
            location_set_map = get_location_set_map(index, index_set)
        self.assertIn('normalize_index.flatten', str(cm.exception))
        normalize_index.flatten = True

        index_set = m.b1d2.index_set()
        index = (1, 'a', 1)
        pred_map = {0: m.time, 1: m.d_2, 2: m.d_2}
        location_set_map = get_location_set_map(index, index_set)
        self.assertSameMap(pred_map, location_set_map)

        index_set = m.b2d2.index_set()
        index = (1, 'a', 1, 2)
        pred_map = {0: m.time, 1: m.d_2, 2: m.d_2, 3: m.space}
        location_set_map = get_location_set_map(index, index_set)
        self.assertSameMap(pred_map, location_set_map)

        index_set = m.b3d2.index_set()
        index = (1, 'a', 1, 2, 1)
        pred_map = {0: m.time, 1: m.d_2, 2: m.d_2, 3: m.space, 4: m.time}
        location_set_map = get_location_set_map(index, index_set)
        self.assertSameMap(pred_map, location_set_map)

    def test_multi_dim_replacement(self):
        m = self.model()

        # One two-dimen set
        index_set = m.b0d2.index_set()
        index = ('a', 1)
        location_set_map = get_location_set_map(index, index_set)
        sets = ComponentSet()
        pred_index = ('a', 1)
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        sets = ComponentSet((m.d_2,))
        pred_index = (slice(None), slice(None))
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        # One two-dimen, one one-dimen
        index_set = m.b1d2.index_set()
        index = (1, 'a', 1)
        location_set_map = get_location_set_map(index, index_set)
        sets = ComponentSet()
        pred_index = (1, 'a', 1)
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        sets = ComponentSet((m.d_2,))
        pred_index = (1, slice(None), slice(None))
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        # One two-dimen, two one-dimen
        index_set = m.b2d2.index_set()
        index = (1, 'a', 1, 2)
        location_set_map = get_location_set_map(index, index_set)
        sets = ComponentSet((m.d_2, m.time))
        pred_index = (slice(None), slice(None), slice(None), 2)
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        sets = ComponentSet((m.space,))
        pred_index = (1, 'a', 1, slice(None))
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        # One two-dimen, three one-dimen
        index_set = m.b3d2.index_set()
        index = (1, 'a', 1, 2, 1)
        location_set_map = get_location_set_map(index, index_set)
        sets = ComponentSet((m.time,))
        pred_index = (slice(None), 'a', 1, 2, slice(None))
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

    def test_dimen_none(self):
        m = self.model()

        index_set = m.b0dn.index_set()
        index = ('c', 1, 10)
        pred_map = {0: m.d_none, 1: m.d_none, 2: m.d_none}
        location_set_map = get_location_set_map(index, index_set)
        self.assertSameMap(pred_map, location_set_map)

        index_set = m.b1dn.index_set()
        index = (1, 'c', 1, 10)
        pred_map = {0: m.time, 1: m.d_none, 2: m.d_none, 3: m.d_none}
        location_set_map = get_location_set_map(index, index_set)
        self.assertSameMap(pred_map, location_set_map)

        index_set = m.b1dnd2.index_set()
        index = (1, 'c', 1, 10, 'a', 1)
        pred_map = {
            0: m.time,
            1: m.d_none,
            2: m.d_none,
            3: m.d_none,
            4: m.d_2,
            5: m.d_2,
        }
        location_set_map = get_location_set_map(index, index_set)
        self.assertSameMap(pred_map, location_set_map)

        index_set = m.b2dn.index_set()
        index = (1, 0, 'd', 3)
        pred_map = {0: m.time, 1: m.space, 2: m.d_none, 3: m.d_none}
        location_set_map = get_location_set_map(index, index_set)
        self.assertSameMap(pred_map, location_set_map)

        index_set = m.b2dnd2.index_set()
        index = (1, 'c', 1, 10, 'b', 2, 0)
        pred_map = {
            0: m.time,
            1: m.d_none,
            2: m.d_none,
            3: m.d_none,
            4: m.d_2,
            5: m.d_2,
            6: m.space,
        }
        location_set_map = get_location_set_map(index, index_set)
        self.assertSameMap(pred_map, location_set_map)

        index_set = m.dnd2b1.index_set()
        index = ('d', 3, 'b', 2, 1)
        pred_map = {0: m.d_none, 1: m.d_none, 2: m.d_2, 3: m.d_2, 4: m.time}
        location_set_map = get_location_set_map(index, index_set)
        self.assertSameMap(pred_map, location_set_map)

        index_set = m.b3dn.index_set()
        index = (1, 'a', 1, 'd', 3, 0, 'b', 2)
        pred_map = {
            0: m.time,
            1: m.d_2,
            2: m.d_2,
            3: m.d_none,
            4: m.d_none,
            5: m.space,
            6: m.d_2,
            7: m.d_2,
        }
        location_set_map = get_location_set_map(index, index_set)
        self.assertSameMap(pred_map, location_set_map)

        index_set = m.dn2.index_set()
        index = (1, 'c', 1, 10, 'b', 2, 'd', 3, 'a', 1)
        with self.assertRaises(RuntimeError) as cm:
            location_set_map = get_location_set_map(index, index_set)
        self.assertIn('multiple sets of dimen==None', str(cm.exception))

    def test_dimen_none_replacement(self):
        m = self.model()

        # Single set of dimen None
        index_set = m.b0dn.index_set()
        index = ('c', 1, 10)
        location_set_map = get_location_set_map(index, index_set)
        sets = ComponentSet((m.d_none,))
        pred_index = (Ellipsis,)
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        sets = ComponentSet()
        pred_index = ('c', 1, 10)
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        # One None-dimen, one one-dimen
        index_set = m.b1dn.index_set()
        index = (1, 'c', 1, 10)
        location_set_map = get_location_set_map(index, index_set)
        sets = ComponentSet((m.d_none,))
        pred_index = (1, Ellipsis)
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        sets = ComponentSet((m.d_none, m.time))
        pred_index = (slice(None), Ellipsis)
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        # One None-dimen, one one-dimen, one two-dimen
        index_set = m.b1dnd2.index_set()
        index = (1, 'c', 1, 10, 'a', 1)
        location_set_map = get_location_set_map(index, index_set)
        sets = ComponentSet((m.d_none, m.time, m.d_2))
        pred_index = (slice(None), Ellipsis, slice(None), slice(None))
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        sets = ComponentSet((m.time, m.d_2))
        pred_index = (slice(None), 'c', 1, 10, slice(None), slice(None))
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        sets = ComponentSet((m.d_2,))
        pred_index = (1, 'c', 1, 10, slice(None), slice(None))
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        # One None-dimen, one two-dimen
        index_set = m.b2dn.index_set()
        index = (1, 0, 'd', 3)
        location_set_map = get_location_set_map(index, index_set)
        sets = ComponentSet()
        pred_index = (1, 0, 'd', 3)
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        sets = ComponentSet((m.d_none,))
        pred_index = (1, 0, Ellipsis)
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        # One None-dimen, one two-dimen, two one-dimen
        index_set = m.b2dnd2.index_set()
        index = (1, 'c', 1, 10, 'b', 2, 0)
        location_set_map = get_location_set_map(index, index_set)
        sets = ComponentSet((m.d_none, m.space))
        pred_index = (1, Ellipsis, 'b', 2, slice(None))
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        # None-dimen set comes first
        index_set = m.dnd2b1.index_set()
        index = ('d', 3, 'b', 2, 1)
        location_set_map = get_location_set_map(index, index_set)
        sets = ComponentSet((m.d_none, m.time))
        pred_index = (Ellipsis, 'b', 2, slice(None))
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

        # None-dimen, plus 2-d set shows up twice
        index_set = m.b3dn.index_set()
        index = (1, 'a', 1, 'd', 3, 0, 'b', 2)
        location_set_map = get_location_set_map(index, index_set)
        sets = ComponentSet((m.d_none, m.d_2))
        pred_index = (
            1,
            slice(None),
            slice(None),
            Ellipsis,
            0,
            slice(None),
            slice(None),
        )
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)

    def test_replacement_normalize_flatten_false(self):
        m = self.model()

        # index of m.b2d2
        index = (1, 0, ('a', 1))
        location_set_map = {0: m.time, 1: m.space, 2: m.d_2}
        sets = ComponentSet((m.d_2,))
        pred_index = (1, 0, slice(None))
        # If you can assemble the proper location_set_map
        # (get_location_set_map will not work), this function
        # should work when normalize_index.flatten is False.
        normalize_index.flatten = False
        new_index = replace_indices(index, location_set_map, sets)
        self.assertEqual(new_index, pred_index)
        normalize_index.flatten = True


class SliceComponentDataAlongSets(unittest.TestCase):
    def model(self):
        m = pyo.ConcreteModel()

        m.time = pyo.Set(initialize=[1, 2, 3])
        m.space = dae.ContinuousSet(initialize=[0, 2])
        m.comp = pyo.Set(initialize=['a', 'b'])
        m.d_2 = pyo.Set(initialize=[('a', 1), ('b', 2)])
        m.d_none = pyo.Set(initialize=[('c', 1, 10), ('d', 3)], dimen=None)

        @m.Block()
        def b(b):
            b.v0 = pyo.Var()
            b.v1 = pyo.Var(m.time)
            b.v2 = pyo.Var(m.time, m.space)

            @b.Block(m.time, m.space)
            def b2(b2):
                b2.v0 = pyo.Var()
                b2.v1 = pyo.Var(m.comp)
                b2.v2 = pyo.Var(m.time, m.comp)
                b2.vn = pyo.Var(m.time, m.d_none, m.d_2)

            @b.Block(m.d_none, m.d_2)
            def bn(bn):
                bn.v0 = pyo.Var()
                bn.v2 = pyo.Var(m.time, m.space)
                bn.v3 = pyo.Var(m.time, m.space, m.time)
                bn.vn = pyo.Var(m.time, m.d_none, m.d_2)

        return m

    def test_with_tuple_of_sets(self):
        m = pyo.ConcreteModel()
        m.s1 = pyo.Set(initialize=[1, 2, 3])
        m.s2 = pyo.Set(initialize=[1, 2, 3])
        m.v = pyo.Var(m.s1, m.s2)
        sets = (m.s1,)
        slice_ = slice_component_along_sets(m.v[1, 2], sets)

        # These tests are essentially the same, but we run the CUID test
        # first as it gives a nicer error message.
        self.assertEqual(str(pyo.ComponentUID(slice_)), "v[*,2]")
        self.assertEqual(slice_, m.v[:, 2])

    def test_no_context(self):
        m = self.model()

        comp = m.b.v0
        sets = ComponentSet((m.time, m.space))
        _slice = slice_component_along_sets(comp, sets)
        # Use `assertIs` when the "slice" is actually just a component.
        self.assertIs(_slice, m.b.v0)

        comp = m.b.v1[1]
        sets = ComponentSet((m.time, m.space))
        _slice = slice_component_along_sets(comp, sets)
        self.assertEqual(_slice, m.b.v1[:])

        comp = m.b.v2[1, 0]
        sets = ComponentSet((m.time, m.space))
        _slice = slice_component_along_sets(comp, sets)
        self.assertEqual(_slice, m.b.v2[:, :])

        comp = m.b.b2[1, 0].v1['a']
        sets = ComponentSet((m.time, m.space))
        _slice = slice_component_along_sets(comp, sets)
        self.assertEqual(_slice, m.b.b2[:, :].v1['a'])

        comp = m.b.b2[1, 0].v1
        sets = ComponentSet((m.time, m.space))
        _slice = slice_component_along_sets(comp, sets)
        self.assertEqual(_slice, m.b.b2[:, :].v1)

        comp = m.b.b2[1, 0].v2[1, 'a']
        sets = ComponentSet((m.time,))
        _slice = slice_component_along_sets(comp, sets)
        self.assertEqual(_slice, m.b.b2[:, 0].v2[:, 'a'])

        comp = m.b.bn['c', 1, 10, 'a', 1].v3[1, 0, 1]
        sets = ComponentSet((m.time,))
        _slice = slice_component_along_sets(comp, sets)
        self.assertEqual(_slice, m.b.bn['c', 1, 10, 'a', 1].v3[:, 0, :])

    def test_context(self):
        m = self.model()

        comp = m.b.v2[1, 0]
        context = m.b
        sets = ComponentSet((m.time,))
        _slice = slice_component_along_sets(comp, sets)
        # Context makes no difference in resulting slice here.
        self.assertEqual(_slice, m.b.v2[:, 0])

        comp = m.b.b2[1, 0].v2[1, 'a']
        context = m.b.b2[1, 0]
        sets = ComponentSet((m.time,))
        _slice = slice_component_along_sets(comp, sets, context=context)
        # Here, context does make a difference
        self.assertEqual(_slice, m.b.b2[1, 0].v2[:, 'a'])

        comp = m.b.b2[1, 0].v1['a']
        context = m.b.b2[1, 0]
        sets = ComponentSet((m.time,))
        _slice = slice_component_along_sets(comp, sets, context=context)
        self.assertIs(_slice, m.b.b2[1, 0].v1['a'])

        sets = ComponentSet((m.comp,))
        _slice = slice_component_along_sets(comp, sets, context=context)
        self.assertEqual(_slice, m.b.b2[1, 0].v1[:])

        context = m.b.b2
        sets = ComponentSet((m.time, m.comp))
        _slice = slice_component_along_sets(comp, sets, context=context)
        self.assertEqual(_slice, m.b.b2[:, 0].v1[:])

    def test_none_dimen(self):
        m = self.model()

        comp = m.b.b2[1, 0].vn
        sets = ComponentSet((m.d_none,))
        _slice = slice_component_along_sets(comp, sets)
        self.assertIs(_slice, m.b.b2[1, 0].vn)

        comp = m.b.b2[1, 0].vn[1, 'd', 3, 'a', 1]
        sets = ComponentSet((m.d_2, m.time))
        _slice = slice_component_along_sets(comp, sets)
        self.assertEqual(_slice, m.b.b2[:, 0].vn[:, 'd', 3, :, :])

        sets = ComponentSet((m.d_none, m.d_2))
        _slice = slice_component_along_sets(comp, sets)
        self.assertEqual(_slice, m.b.b2[1, 0].vn[1, ..., :, :])

        comp = m.b.bn['c', 1, 10, 'a', 1].v3[1, 0, 1]
        sets = ComponentSet((m.d_none, m.time))
        _slice = slice_component_along_sets(comp, sets)
        self.assertEqual(_slice, m.b.bn[..., 'a', 1].v3[:, 0, :])

        comp = m.b.bn['c', 1, 10, 'a', 1].vn[1, 'd', 3, 'b', 2]
        sets = ComponentSet((m.d_none,))
        _slice = slice_component_along_sets(comp, sets)
        self.assertEqual(_slice, m.b.bn[..., 'a', 1].vn[1, ..., 'b', 2])

        sets = ComponentSet((m.d_none, m.d_2))
        _slice = slice_component_along_sets(comp, sets)
        self.assertEqual(_slice, m.b.bn[..., :, :].vn[1, ..., :, :])

        comp = m.b.bn['c', 1, 10, 'a', 1].vn[1, 'd', 3, 'b', 2]
        context = m.b.bn['c', 1, 10, 'a', 1]
        sets = ComponentSet((m.d_none,))
        _slice = slice_component_along_sets(comp, sets, context=context)
        self.assertEqual(_slice, m.b.bn['c', 1, 10, 'a', 1].vn[1, ..., 'b', 2])

        sets = ComponentSet((m.d_none, m.d_2))
        _slice = slice_component_along_sets(comp, sets, context=context)
        self.assertEqual(_slice, m.b.bn['c', 1, 10, 'a', 1].vn[1, ..., :, :])

        context = m.b.bn
        sets = ComponentSet((m.d_none,))
        _slice = slice_component_along_sets(comp, sets, context=context)
        self.assertEqual(_slice, m.b.bn[..., 'a', 1].vn[1, ..., 'b', 2])


class TestCUID(unittest.TestCase):
    def test_cuid_of_slice(self):
        m = pyo.ConcreteModel()
        m.s1 = pyo.Set(initialize=["a", "b"])
        m.s2 = pyo.Set(initialize=["c", "d"])
        m.b = pyo.Block(m.s1)
        for i in m.s1:
            m.b[i].v = pyo.Var(m.s2)
        slice_ = slice_component_along_sets(m.b["a"].v["c"], ComponentSet((m.s1,)))
        cuid = pyo.ComponentUID(slice_)
        self.assertEqual(str(cuid), "b[*].v[c]")

        slice_ = slice_component_along_sets(m.b["a"].v[("c",)], ComponentSet((m.s1,)))
        cuid = pyo.ComponentUID(slice_)
        self.assertEqual(str(cuid), "b[*].v[c]")


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.config import ConfigDict, ConfigValue
from pyomo.common.errors import DeveloperError
from pyomo.common.modeling import unique_component_name
from pyomo.contrib.piecewise import PiecewiseLinearFunction
from pyomo.contrib.piecewise.transform.piecewise_to_mip_visitor import (
    PiecewiseLinearToMIP,
)
from pyomo.core import (
    Constraint,
    Objective,
    Var,
    BooleanVar,
    Expression,
    Suffix,
    Param,
    Set,
    SetOf,
    RangeSet,
    ExternalFunction,
    Connector,
    SortComponents,
    Any,
    LogicalConstraint,
)
from pyomo.core.base import Transformation
from pyomo.core.base.block import Block
from pyomo.core.util import target_list
from pyomo.gdp import Disjunct, Disjunction
from pyomo.gdp.util import is_child_of
from pyomo.network import Port


class PiecewiseLinearTransformationBase(Transformation):
    """
    Base class for transformations of piecewise-linear models to GDPs, MIPs, etc.
    """

    CONFIG = ConfigDict('contrib.piecewise_linear_transformation_base')
    CONFIG.declare(
        'targets',
        ConfigValue(
            default=None,
            domain=target_list,
            description="target or list of targets that will be transformed",
            doc="""
            This specifies the list of components to transform. If None (default),
            the entire model is transformed. Note that if the transformation is
            done out of place, the list of targets should be attached to the model
            before it is cloned, and the list will specify the targets on the cloned
            instance.""",
        ),
    )
    CONFIG.declare(
        'descend_into_expressions',
        ConfigValue(
            default=False,
            domain=bool,
            description="Whether to look for uses of PiecewiseLinearFunctions in "
            "the Constraint and Objective expressions, rather than assuming "
            "all PiecewiseLinearFunctions are on the active tree(s) of 'instance' "
            "and 'targets.'",
            doc="""
            It is *strongly* recommended that, in hierarchical models, the
            PiecewiseLinearFunction components are on the same Block as where
            they are used in expressions. If you follow this recommendation,
            this option can remain False, which will make this transformation
            more efficient. However, if you do not follow the recommendation,
            unless you know what you are doing, turn this option to 'True' to
            ensure that all of the uses of PiecewiseLinearFunctions are
            transformed.
            """,
        ),
    )

    def __init__(self):
        super().__init__()
        self.handlers = {
            Constraint: self._transform_constraint,
            Objective: self._transform_objective,
            Var: False,
            BooleanVar: False,
            Connector: False,
            Expression: False,
            Suffix: False,
            Param: False,
            Set: False,
            SetOf: False,
            RangeSet: False,
            Disjunction: False,
            Disjunct: False,
            Block: self._transform_block,
            ExternalFunction: False,
            Port: False,
            PiecewiseLinearFunction: self._transform_piecewise_linear_function,
            LogicalConstraint: False,
        }
        self._transformation_blocks = {}

    def _apply_to(self, instance, **kwds):
        try:
            self._apply_to_impl(instance, **kwds)
        finally:
            self._transformation_blocks.clear()

    def _apply_to_impl(self, instance, **kwds):
        config = self.CONFIG(kwds.pop('options', {}))
        config.set_value(kwds)

        targets = config.targets
        if targets is None:
            targets = (instance,)

        knownBlocks = {}
        not_walking_exprs_msg = (
            "When not descending into expressions, Constraints "
            "and Objectives are not valid targets. Please specify "
            "PiecewiseLinearFunction component and the Blocks "
            "containing them, or (at the cost of some performance "
            "in this transformation), set the 'descend_into_expressions' "
            "option to 'True'."
        )
        for t in targets:
            if not is_child_of(parent=instance, child=t, knownBlocks=knownBlocks):
                raise ValueError(
                    "Target '%s' is not a component on instance "
                    "'%s'!" % (t.name, instance.name)
                )
            if t.ctype is PiecewiseLinearFunction:
                if config.descend_into_expressions:
                    raise ValueError(
                        "When descending into expressions, the transformation "
                        "cannot take PiecewiseLinearFunction components as "
                        "targets. Please instead specify the Blocks, "
                        "Constraints, and Objectives where your "
                        "PiecewiseLinearFunctions have been used in "
                        "expressions."
                    )
                self._transform_piecewise_linear_function(
                    t, config.descend_into_expressions
                )
            elif issubclass(t.ctype, Block):
                self._transform_block(t, config.descend_into_expressions)
            elif t.ctype is Constraint:
                if not config.descend_into_expressions:
                    raise ValueError(
                        "Encountered Constraint target '%s':\n%s"
                        % (t.name, not_walking_exprs_msg)
                    )
                self._transform_constraint(t, config.descend_into_expressions)
            elif t.ctype is Objective:
                if not config.descend_into_expressions:
                    raise ValueError(
                        "Encountered Objective target '%s':\n%s"
                        % (t.name, not_walking_exprs_msg)
                    )
                self._transform_objective(t, config.descend_into_expressions)
            else:
                raise ValueError(
                    "Target '%s' is not a PiecewiseLinearFunction, Block or "
                    "Constraint. It was of type '%s' and can't be transformed."
                    % (t.name, type(t))
                )

    def _get_transformation_block(self, parent):
        if parent in self._transformation_blocks:
            return self._transformation_blocks[parent]

        nm = unique_component_name(
            parent, '_pyomo_contrib_%s' % self._transformation_name
        )
        self._transformation_blocks[parent] = transBlock = Block()
        parent.add_component(nm, transBlock)

        transBlock.transformed_functions = Block(Any)
        return transBlock

    def _transform_block(self, block, descend_into_expressions):
        blocks = block.values() if block.is_indexed() else (block,)
        for b in blocks:
            for obj in b.component_objects(
                active=True,
                descend_into=(Block, Disjunct),
                sort=SortComponents.deterministic,
            ):
                handler = self.handlers.get(obj.ctype, None)
                if not handler:
                    if handler is None:
                        raise RuntimeError(
                            "No transformation handler registered for modeling "
                            "components of type '%s'." % obj.ctype
                        )
                    continue
                handler(obj, descend_into_expressions)

    def _transform_piecewise_linear_function(
        self, pw_linear_func, descend_into_expressions
    ):
        if descend_into_expressions:
            return

        transBlock = self._get_transformation_block(pw_linear_func.parent_block())
        _functions = (
            pw_linear_func.values()
            if pw_linear_func.is_indexed()
            else (pw_linear_func,)
        )
        for pw_func in _functions:
            for pw_expr in pw_func._expressions.values():
                substitute_var = self._transform_pw_linear_expr(
                    pw_expr.expr, pw_func, transBlock
                )
                # We change the named expression to point to the variable that
                # will take the appropriate value of the piecewise linear
                # function.
                pw_expr.expr = substitute_var

        # Deactivate so that modern writers don't complain
        pw_linear_func.deactivate()

    def _transform_constraint(self, constraint, descend_into_expressions):
        if not descend_into_expressions:
            return

        transBlock = self._get_transformation_block(constraint.parent_block())
        visitor = PiecewiseLinearToMIP(self._transform_pw_linear_expr, transBlock)

        _constraints = constraint.values() if constraint.is_indexed() else (constraint,)
        for c in _constraints:
            visitor.walk_expression((c.expr, c, 0))

    def _transform_objective(self, objective, descend_into_expressions):
        if not descend_into_expressions:
            return

        transBlock = self._get_transformation_block(objective.parent_block())
        visitor = PiecewiseLinearToMIP(self._transform_pw_linear_expr, transBlock)

        _objectives = objective.values() if objective.is_indexed() else (objective,)
        for o in _objectives:
            visitor.walk_expression((o.expr, o, 0))

    def _transform_pw_linear_expr(self, pw_expr, pw_linear_func, transformation_block):
        raise DeveloperError(
            "Derived class failed to implement '_transform_pw_linear_expr'"
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# The parametric.mod example from sIPOPT
#
# Original implementation by Hans Pirnay is in pyomo/examples/pyomo/suffixes
#

from pyomo.environ import *
from pyomo.contrib.sensitivity_toolbox.sens import sensitivity_calculation


def create_model():
    '''Create a concrete Pyomo model for this example'''
    m = ConcreteModel()

    m.x1 = Var(initialize=0.15, within=NonNegativeReals)
    m.x2 = Var(initialize=0.15, within=NonNegativeReals)
    m.x3 = Var(initialize=0.0, within=NonNegativeReals)

    m.eta1 = Param(initialize=4.5, mutable=True)
    m.eta2 = Param(initialize=1.0, mutable=True)

    m.const1 = Constraint(expr=6 * m.x1 + 3 * m.x2 + 2 * m.x3 - m.eta1 == 0)
    m.const2 = Constraint(expr=m.eta2 * m.x1 + m.x2 - m.x3 - 1 == 0)
    m.cost = Objective(expr=m.x1**2 + m.x2**2 + m.x3**2)

    return m


def run_example(print_flag=True):
    '''
    Execute the example

    Arguments:
        print_flag: Toggle on/off printing

    Returns
        sln_dict: Dictionary containing solution (used for automated testing)

    '''
    m = create_model()

    m.perturbed_eta1 = Param(initialize=4.0)
    m.perturbed_eta2 = Param(initialize=1.0)

    m_kaug_dsdp = sensitivity_calculation(
        'k_aug', m, [m.eta1, m.eta2], [m.perturbed_eta1, m.perturbed_eta2], tee=True
    )

    if print_flag:
        print("\nOriginal parameter values:")
        print("\teta1 =", m.eta1())
        print("\teta2 =", m.eta2())

        print("Initial point:")
        print("\tObjective =", value(m.cost))
        print("\tx1 =", m.x1())
        print("\tx2 =", m.x2())
        print("\tx3 =", m.x3())

        # Kaug saves only approximated solutions not original solutions
        print("\nNew parameter values:")
        print("\teta1 =", m_kaug_dsdp.perturbed_eta1())
        print("\teta2 =", m_kaug_dsdp.perturbed_eta2())

        print("(Approximate) solution with the new parameter values:")
        print("\tObjective =", m_kaug_dsdp.cost())
        print("\tx1 =", m_kaug_dsdp.x1())
        print("\tx2 =", m_kaug_dsdp.x2())
        print("\tx3 =", m_kaug_dsdp.x3())

    # Save the results in a dictionary.
    # This is optional and makes automated testing convenient.
    # This code is not required for a Minimum Working Example (MWE)
    d = dict()
    d['eta1'] = m.eta1()
    d['eta2'] = m.eta2()
    d['x1_init'] = m.x1()
    d['x2_init'] = m.x2()
    d['x3_init'] = m.x3()
    d['eta1_pert'] = m_kaug_dsdp.perturbed_eta1()
    d['eta2_pert'] = m_kaug_dsdp.perturbed_eta2()
    d['cost_pert'] = m_kaug_dsdp.cost()
    d['x1_pert'] = m_kaug_dsdp.x1()
    d['x2_pert'] = m_kaug_dsdp.x2()
    d['x3_pert'] = m_kaug_dsdp.x3()

    return d


if __name__ == '__main__':
    d = run_example()
